import re
from typing import List, Optional

from pydantic import BaseModel, validator

from opennem.core.normalizers import normalize_whitespace


class SchemaBase(BaseModel):
    class Config:
        orm_mode = True
        anystr_strip_whitespace = True


class FacilitySchema(SchemaBase):
    network_region: str
    status: str
    duid: str
    name: str
    fueltech: Optional[str] = None
    capacity: Optional[float] = None


class StationSchema(SchemaBase):
    name: str
    code: Optional[str] = None
    state: Optional[str] = None
    facilities: List[FacilitySchema] = []

    @classmethod
    @validator("name")
    def name_clean(cls, v):
        name = normalize_whitespace(v)

        return name

    @classmethod
    @validator("state")
    def state_clean(cls, v):
        state = state.upper()

        return state



from datetime import date, datetime
from typing import List, Optional

from opennem.api.schema import ApiBase


class WeatherObservation(ApiBase):
    observation_time: datetime
    station_id: int

    temp_apparent: float
    temp_air: float
    press_qnh: float
    wind_dir: Optional[str]
    wind_spd: Optional[float]
    wind_gust: Optional[float]
    humidity: Optional[float]
    cloud: Optional[str]
    cloud_type: Optional[str]


class WeatherStation(ApiBase):
    code: str
    state: str
    name_alias: str
    registered: date
    website_url: str
    altitude: int
    lat: float
    lng: float

    observations: Optional[List[WeatherObservation]]



import logging

from sqlalchemy.dialects.postgresql import insert

from opennem.db import SessionLocal, get_database_engine
from opennem.db.models.opennem import BomObservation
from opennem.utils.dates import parse_date
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


class StoreBomObservation(object):
    """
        Pipeline to store BOM observations into the database
    """

    @check_spider_pipeline
    def process_item(self, item, spider):

        session = SessionLocal()
        engine = get_database_engine()

        records_to_store = []

        if "records" not in item:

            return 0
        records = item["records"]

        for obs in records:
            observation_time = parse_date(
                obs["aifstime_utc"], dayfirst=False, is_utc=True
            )
            code = obs["code"]

            if not observation_time or not code:
                continue

            records_to_store.append(
                {
                    "station_id": code,
                    "observation_time": observation_time,
                    "temp_apparent": obs["apparent_t"],
                    "temp_air": obs["air_temp"],
                    "press_qnh": obs["press_qnh"],
                    "wind_dir": obs["wind_dir"],
                    "wind_spd": obs["wind_spd_kmh"],
                    "wind_gust": obs["gust_kmh"],
                    "cloud": obs["cloud"].replace("-", ""),
                    "cloud_type": obs["cloud_type"].replace("-", ""),
                    "humidity": obs["rel_hum"],
                }
            )

        if not len(records_to_store):
            return 0

        stmt = insert(BomObservation).values(records_to_store)
        stmt.bind = engine
        stmt = stmt.on_conflict_do_update(
            index_elements=["observation_time", "station_id"],
            set_={
                "temp_apparent": stmt.excluded.temp_apparent,
                "temp_air": stmt.excluded.temp_air,
                "press_qnh": stmt.excluded.press_qnh,
                "wind_dir": stmt.excluded.wind_dir,
                "wind_spd": stmt.excluded.wind_spd,
                "wind_gust": stmt.excluded.wind_gust,
                "cloud": stmt.excluded.cloud,
                "cloud_type": stmt.excluded.cloud_type,
                "humidity": stmt.excluded.humidity,
            },
        )

        try:
            session.execute(stmt)
            session.commit()
        except Exception as e:
            logger.error("Error: {}".format(e))
            return 0
        finally:
            session.close()

        return len(records_to_store)



from datetime import datetime
from typing import List

from fastapi import APIRouter, Depends, HTTPException, Query
from sqlalchemy.orm import Session
from starlette import status

from opennem.db import get_database_session
from opennem.db.models.opennem import Facility

from .schema import (
    FacilityModification,
    FacilityModificationTypes,
    FacilityRecord,
    FacilityUpdateResponse,
)

router = APIRouter()


@router.get(
    "/",
    name="Facilities",
    description="Get facilities",
    response_model=List[FacilityRecord],
)
def facilities(
    session: Session = Depends(get_database_session),
) -> FacilityRecord:
    facilities = session.query(Facility).all()

    return facilities


@router.get(
    "/{facility_code:path}", name="Facility", response_model=FacilityRecord,
)
def facility(
    facility_code: str = Query(str, description="Facility code"),
    session: Session = Depends(get_database_session),
) -> FacilityRecord:
    facility = (
        session.query(Facility).filter_by(code=facility_code).one_or_none()
    )

    if not facility:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Facility not found"
        )

    return facility


@router.put("/{facility_id}", name="Facility update")
def facility_update(
    facility_id: int,
    data: FacilityModification = {},
    session: Session = Depends(get_database_session),
) -> dict:
    facility = session.query(Facility).get(facility_id)

    if not facility:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Facility not found"
        )

    if data.modification == FacilityModificationTypes.approve:

        if facility.approved is True:
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail="Facility already approved",
            )

        facility.approved = True
        facility.approved_at = datetime.now()
        facility.approved_by = "opennem.admin"

    if data.modification == FacilityModificationTypes.reject:
        facility.approved = False

    session.add(facility)
    session.commit()

    response = FacilityUpdateResponse(success=True, record=facility)

    return response



import csv
import logging
from io import StringIO

from sqlalchemy.sql.schema import Table

from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


def generate_csv_from_records(table: Table, records):
    csv_buffer = StringIO()

    column_names = [c.name for c in table.__table__.columns.values()]

    csvwriter = csv.DictWriter(csv_buffer, fieldnames=column_names)

    for record in records:
        csvwriter.writerow(record)

    csv_buffer.seek(0)
    return csv_buffer


class RecordsToCSVPipeline(object):
    @check_spider_pipeline
    def process_item(self, item, spider=None):
        if not isinstance(item, dict):
            logger.error("Invalid item passed to CSV pipeline: %s", item)
            return item

        if "table_schema" not in item:
            logger.error("No table model passed to csv generator")
            logger.error(item)
            return item

        table = item["table_schema"]

        if "records" not in item:
            logger.error("No records to generate csv from")
            return item

        records = item["records"]

        csv_content = generate_csv_from_records(table, records)

        item["csv"] = csv_content

        return item



from collections import OrderedDict
from datetime import datetime, timezone
from itertools import groupby
from operator import attrgetter
from typing import List, Optional, Union

from pytz import timezone as pytz_timezone
from sqlalchemy.orm import Session

from opennem.db.models.opennem import FacilityScada, Station
from opennem.schema.network import NetworkSchema
from opennem.schema.time import TimeInterval, TimePeriod
from opennem.schema.units import UnitDefinition
from opennem.utils.time import human_to_timedelta
from opennem.utils.timezone import is_aware, make_aware

from .schema import (
    DataQueryResult,
    OpennemData,
    OpennemDataHistory,
    OpennemDataSet,
)


def stats_factory(
    stats: List[DataQueryResult],
    interval: TimeInterval,
    period: TimePeriod,
    units: UnitDefinition,
    network: Optional[NetworkSchema] = None,
    timezone: Optional[Union[timezone, pytz_timezone]] = None,
    code: Optional[str] = None,
    region: Optional[str] = None,
    fueltech_group: bool = False,
    group_field: str = None,
) -> Optional[OpennemDataSet]:
    """
        Takes a list of data query results and returns OpennemDataSets

        @TODO optional groupby field
        @TODO override timezone
    """

    if network:
        timezone = network.get_timezone()

    dates = [s.interval for s in stats]

    start = min(dates)
    end = max(dates)

    # should probably make sure these are the same TZ
    if timezone and not is_aware(start):
        start = make_aware(min(dates), timezone)

    if timezone and not is_aware(end):
        end = make_aware(max(dates), timezone)

    # free
    dates = []

    group_codes = list(set([i.group_by for i in stats if i.group_by]))

    stats_grouped = []

    for group_code in group_codes:

        data_grouped = dict()

        for stat in stats:
            if stat.interval not in data_grouped:
                data_grouped[stat.interval] = None

            if stat.group_by == group_code:
                data_grouped[stat.interval] = stat.result

        data_sorted = OrderedDict(sorted(data_grouped.items()))

        history = OpennemDataHistory(
            start=start,
            last=end,
            interval=interval.interval_human,
            data=list(data_sorted.values()),
        )

        data = OpennemData(
            data_type=units.unit_type,
            units=units.unit,
            code=group_code,
            interval=interval,
            period=period,
            history=history,
        )

        if network:
            data.network = network.code

        # *sigh* - not the most flexible model
        if fueltech_group:
            data.fuel_tech = group_code
            data.id = ".".join(
                [network.code, "fueltech", group_code, units.unit_type]
            )
            # @TODO make this an alias
            data.type = units.unit_type

        if group_field:
            group_fields = []

            # setattr(data, group_field, group_code)

            if network:
                group_fields.append(network.code)

            group_fields = group_fields + [
                group_field,
                group_code,
                units.unit_type,
            ]

            data.id = ".".join([f for f in group_fields if f])
            data.type = units.unit_type

        if region:
            data.region = region

        stats_grouped.append(data)

    stat_set = OpennemDataSet(
        data_type=units.unit_type,
        data=stats_grouped,
        created_at=datetime.now(),
        code=code,
    )

    if network:
        stat_set.network = network.code

    if region:
        stat_set.region = region

    return stat_set


def station_attach_stats(station: Station, session: Session) -> Station:
    # @TODO update for new queries
    since = datetime.now() - human_to_timedelta("7d")

    facility_codes = list(set([f.code for f in station.facilities]))

    stats = (
        session.query(FacilityScada)
        .filter(FacilityScada.facility_code.in_(facility_codes))
        .filter(FacilityScada.trading_interval >= since)
        .order_by(FacilityScada.facility_code)
        .order_by(FacilityScada.trading_interval)
        .all()
    )

    for facility in station.facilities:
        facility_power = list(
            filter(lambda s: s.facility_code == facility.code, stats)
        )

        # facility.scada_power = stats_factory(facility_power)

    return station



import logging
import re
from datetime import datetime

import scrapy
from scrapy import Spider

from opennem.utils.dates import parse_date

PADDING_WIDTH = 3

__is_number = re.compile(r"^\d+$")


def is_number(value):
    if re.match(__is_number, value):
        return True
    return False


def parse_dirlisting(raw_string):
    """
        given a raw text directory listing like "     Saturday 11th June 2020      6789"
        will parse and return both the date and listing type

        @param raw_string - the raw directory listing string
        @return dict of the date in iso format and the type (file or directory)
    """
    components = raw_string.split(" " * PADDING_WIDTH)
    components = [i.strip() for i in components]
    components = list(filter(lambda x: x != "", components))

    _ltype = "dir"

    if not components or len(components) < 2:
        logging.debug(components)
        raise Exception(
            "Invalid line string: {}. Components are: {}".format(
                raw_string, components
            )
        )

    if is_number(components[1]):
        _ltype = "file"

    dt = parse_date(components[0])

    if type(dt) is not datetime:
        raise Exception(
            "{} is not a valid datetime. Original value was {}".format(
                dt, components[0]
            )
        )

    return {
        "date": dt.isoformat(),
        "type": _ltype,
    }


class DirlistingSpider(Spider):
    """
        spider that parses html directory listings produced by web servers


    """

    limit = 0

    skip = 0

    start_url = None

    supported_extensions = [".csv", ".zip"]

    def start_requests(self):
        if self.custom_settings is None:
            self.custom_settings = {}

        starts = None

        if (
            hasattr(self, "start_url")
            and self.start_url
            and type(self.start_url) is str
        ):
            starts = [self.start_url]

        if (
            hasattr(self, "start_urls")
            and self.start_urls
            and type(self.start_urls) is list
        ):
            starts = [self.start_urls]

        for url in starts:
            yield scrapy.Request(url)

    def parse(self, response):
        links = list(
            reversed(
                [
                    i.get()
                    for i in response.xpath(
                        "//body/pre/br/following-sibling::a/@href"
                    )
                ]
            )
        )

        metadata = list(
            reversed(
                [
                    parse_dirlisting(i.get())
                    for i in response.xpath(
                        "//body/pre/br/following-sibling::text()"
                    )
                ]
            )
        )

        parsed = 0

        for i, entry in enumerate(metadata):
            if entry["type"] == "file":
                link = response.urljoin(links[i])

                self.log("Getting {}".format(link), logging.INFO)

                if self.limit and self.limit > 0 and (parsed >= self.limit):
                    self.log(f"Reached limit of {self.limit}", logging.INFO)
                    return None

                parsed += 1

                if self.skip and self.skip >= parsed:
                    self.log(
                        f"Skipping entry {link} - {parsed}. Skip set to {self.skip}",
                        logging.INFO,
                    )
                    continue

                if link[-4:].lower() in self.supported_extensions:
                    yield {"link": link, **entry}
                else:
                    self.log(
                        "Not a supported extension: {} {}".format(
                            link[-4:].lower(),
                            ", ".join(self.supported_extensions),
                        )
                    )



import csv
import logging
import zipfile

from scrapy.exceptions import DropItem
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker

from opennem.db import db_connect
from opennem.db.models.opennem import DispatchPrice, DispatchUnitScada
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)

TABLE_MAP = {
    "DISPATCH_UNIT_SCADA": DispatchUnitScada,
    "DISPATCH_PRICE": DispatchPrice,
}


class DatabaseStore(object):
    def __init__(self):
        engine = db_connect()
        self.session = sessionmaker(bind=engine)

    @check_spider_pipeline
    def process_item(self, item, spider):

        s = self.session()

        if not "tables" in item:
            logger.info("No tables in item to process")
            return item

        tables = item["tables"]

        for table in tables:

            if not "name" in table:
                logger.error("No name in table")
                continue

            table_name = table["name"]

            if not table_name in TABLE_MAP:
                logger.error(
                    "Could not map table name to model for item {}".format(
                        table_name
                    )
                )
                continue

            Model = TABLE_MAP[table_name]

            if not "records" in table:
                logger.info(
                    "No records for item with table name {}".format(table_name)
                )
                continue

            records = table["records"]

            for record in records:

                i = Model(**record)

                try:
                    s.add(i)
                    s.commit()
                except IntegrityError:
                    pass
                except Exception as e:
                    s.rollback()
                    logger.error("Database error: {}".format(e))
                    logger.debug(i.__dict__)
                finally:
                    s.close()

        s.close()

        return item



from datetime import datetime
from typing import List, Optional

import sqlalchemy
from fastapi import APIRouter, Depends, HTTPException, Query
from sqlalchemy.orm import Session, joinedload, noload, raiseload
from starlette import status

from opennem.api.facility.schema import FacilityModification
from opennem.core.dispatch_type import DispatchType
from opennem.db import get_database_engine, get_database_session
from opennem.db.models.opennem import (
    BomStation,
    Facility,
    FuelTech,
    Location,
    Network,
    Photo,
    Revision,
    Station,
)
from opennem.schema.opennem import StationSchema

from .schema import (
    StationIDList,
    StationModification,
    StationRecord,
    StationUpdateResponse,
)

router = APIRouter()


@router.get(
    "/",
    response_model=List[StationRecord],
    description="Get a list of all stations",
)
def stations(
    session: Session = Depends(get_database_session),
    facilities_include: Optional[bool] = Query(
        False, description="Include facilities in records"
    ),
    revisions_include: Optional[bool] = Query(
        False, description="Include revisions in records"
    ),
    history_include: Optional[bool] = Query(
        False, description="Include history in records"
    ),
    only_approved: Optional[bool] = Query(
        False, description="Only show approved stations not those pending"
    ),
    name: Optional[str] = None,
    limit: Optional[int] = None,
    page: int = 1,
) -> List[StationSchema]:
    stations = (
        session.query(Station)
        .join(Location)
        .enable_eagerloads(True)
        #
        # .group_by(Station.code)
        # .filter(Facility.fueltech_id.isnot(None))
        # .filter(Facility.status_id.isnot(None))
    )

    if facilities_include:
        stations = stations.outerjoin(
            Facility, Facility.station_id == Station.id
        ).outerjoin(FuelTech, Facility.fueltech_id == FuelTech.code)

    if revisions_include:
        stations = stations.outerjoin(
            Revision, Revision.station_id == Station.id
        )

    if only_approved:
        stations = stations.filter(Station.approved == True)

    if name:
        stations = stations.filter(Station.name.like("%{}%".format(name)))

    stations = stations.order_by(
        # Facility.network_region,
        Station.name,
        # Facility.network_code,
        # Facility.code,
    )

    stations = stations.all()

    return stations


@router.get(
    "/",
    name="station lookup",
    description="""Lookup station by code or network_code. \
        Require one of code or network_code""",
    response_model=List[StationRecord],
)
def station_lookup(
    session: Session = Depends(get_database_session),
    station_code: Optional[str] = Query(
        None, description="Code of the station to lookup"
    ),
    network_code: Optional[str] = Query(
        None, description="The network code to lookup"
    ),
) -> List[StationSchema]:
    if not station_code and not network_code:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    station = session.query(Station)

    if station_code:
        station = station.filter_by(code=station_code)

    if network_code:
        station = station.filter_by(network_name=network_code)

    stations = station.all()

    if not stations:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return stations


@router.get(
    "/ids",
    response_model=List[StationIDList],
    description="Get a list of station ids for dropdowns",
)
def station_ids(
    session: Session = Depends(get_database_session),
    only_approved: Optional[bool] = Query(
        True, description="Only show approved stations not those pending"
    ),
) -> List[StationIDList]:
    stations = session.query(Station).join(Station.location)

    if only_approved:
        stations = stations.filter(Station.approved == True)

    stations = stations.order_by(Station.id,)

    stations = stations.all()

    return stations


@router.get(
    "/{station_code:path}",
    response_model=StationSchema,
    description="Get a single station by code",
)
def station(
    session: Session = Depends(get_database_session),
    engine=Depends(get_database_engine),
    station_code: str = None,
    only_generators: bool = Query(True, description="Show only generators"),
    power_include: Optional[bool] = Query(
        False, description="Include last week of power output"
    ),
    revisions_include: Optional[bool] = Query(
        False, description="Include revisions in records"
    ),
    history_include: Optional[bool] = Query(
        False, description="Include history in records"
    ),
):

    station = (
        session.query(Station)
        .filter(Station.code == station_code)
        .filter(Facility.station_id == Station.id)
        .filter(~Facility.code.endswith("NL1"))
    )

    if only_generators:
        station = station.filter(
            Facility.dispatch_type == DispatchType.GENERATOR
        )

    station = station.one_or_none()

    if not station:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Station not found"
        )

    if revisions_include:
        revisions = session.query(Revision).all()

        station.revisions = list(
            filter(
                lambda rev: rev.schema == "station"
                and rev.code == station.code,
                revisions,
            )
        )

        for facility in station.facilities:
            facility.revisions = list(
                filter(
                    lambda rev: rev.schema == "facility"
                    and rev.code == facility.code,
                    revisions,
                )
            )

    if power_include:
        pass

    if station.location and station.location.geom:
        __query = """
            select
                code,
                ST_Distance(l.geom, bs.geom, false) / 1000.0 as dist
            from bom_station bs, location l
            where
                l.id = {id}
                and bs.priority < 2
            order by dist
            limit 1
        """.format(
            id=station.location.id
        )

        result = []

        with engine.connect() as c:
            result = list(c.execute(__query))

            if len(result):
                station.location.weather_nearest = {
                    "code": result[0][0],
                    "distance": round(result[0][1], 2),
                }

    return station


@router.get(
    "/history",
    name="station all history",
    description="""Get history for all stations""",
    response_model=List[StationSchema],
)
def stations_history(session: Session = Depends(get_database_session)):
    raise HTTPException(status_code=status.HTTP_501_NOT_IMPLEMENTED)


@router.get(
    "/station/history/{station_code}",
    name="station history",
    description="""Get history for a station""",
    response_model=StationSchema,
)
def station_history(
    station_code: str, session: Session = Depends(get_database_session)
):
    raise HTTPException(status_code=status.HTTP_501_NOT_IMPLEMENTED)


@router.post(
    "/",
    name="station",
    # response_model=StationSchema,
    description="Create a station",
)
def station_create(
    session: Session = Depends(get_database_session),
    # station: StationSubmission = None,
):
    raise HTTPException(status_code=status.HTTP_501_NOT_IMPLEMENTED)


@router.put("/{station_id}", name="Station update")
def station_update(
    station_id: int,
    data: StationModification = {},
    session: Session = Depends(get_database_session),
) -> dict:
    station = session.query(Station).get(station_id)

    if not station:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Station not found"
        )

    if data.modification == "approve":

        if station.approved is True:
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail="Station already approved",
            )

        station.approved = True
        station.approved_at = datetime.now()
        station.approved_by = "opennem.admin"

    if data.modification == "reject":
        station.approved = False

    session.add(station)
    session.commit()

    response = StationUpdateResponse(success=True, record=station)

    return response



from datetime import datetime

from fastapi import APIRouter, Depends, HTTPException
from fastapi.param_functions import Query
from starlette import status
from starlette.status import HTTP_501_NOT_IMPLEMENTED

from opennem.api.stats.router import (
    energy_network_fueltech_api,
    power_network_fueltech_api,
    price_network_region_api,
)
from opennem.api.stats.schema import OpennemDataSet
from opennem.api.weather.router import station_observations_api
from opennem.core.networks import network_from_network_code
from opennem.db import get_database_engine
from opennem.schema.network import NetworkSchema

router = APIRouter()

YEAR_CURRENT = datetime.now().date().year


@router.get(
    "/power/wem.json",
    name="Network power",
    response_model_exclude_unset=True,
    response_model=OpennemDataSet,
)
def api_export_power_wem(
    engine=Depends(get_database_engine),
) -> OpennemDataSet:
    stats = power_network_fueltech_api(
        network_code="WEM",
        network_region="WEM",
        interval="30m",
        period="7d",
        engine=engine,
    )

    weather = station_observations_api(
        station_code="009021",
        interval="30m",
        period="7d",
        network_code="WEM",
        engine=engine,
    )

    price = price_network_region_api(
        engine=engine,
        network_code="WEM",
        region_code="WEM",
        interval="30m",
        period="7d",
    )

    # demand = wem_demand()

    stats.data = stats.data + price.data + weather.data

    return stats


@router.get(
    "/{network_code}/energy/daily/{year}.json",
    name="Network energy by year",
    response_model_exclude_unset=True,
    response_model=OpennemDataSet,
)
def api_export_energy_year(
    engine=Depends(get_database_engine),
    network_code: str = Query("WEM", description="Network code"),
    year: int = Query(YEAR_CURRENT, description="Year to query"),
) -> OpennemDataSet:
    if year > YEAR_CURRENT or year < 1996:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Invalid year"
        )

    network: NetworkSchema = network_from_network_code(network_code)

    if not network:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Invalid network",
        )

    stats = energy_network_fueltech_api(
        network_code=network.code,
        network_region="WEM",
        interval="1d",
        year=year,
        period="1Y",
        engine=engine,
    )

    return stats


@router.get(
    "/{network_code}/energy/monthly/{month}.json",
    name="Network energy by month",
    response_model_exclude_unset=True,
    response_model=OpennemDataSet,
)
def api_export_energy_month(
    engine=Depends(get_database_engine),
    network_code: str = Query("WEM", description="Network code"),
    month: str = Query("all", description="Month to query"),
) -> OpennemDataSet:
    network: NetworkSchema = network_from_network_code(network_code)

    if month != "all":
        raise HTTPException(
            status_code=status.HTTP_501_NOT_IMPLEMENTED,
            detail="Other months not yet supported",
        )

    if not network:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Invalid network",
        )

    stats = energy_network_fueltech_api(
        network_code=network.code,
        network_region="WEM",
        interval="1M",
        period="10Y",
        engine=engine,
    )

    return stats



"""Updated indexes on balancing summary and facility scada

Revision ID: 9ee998713b18
Revises: 3595c70244db
Create Date: 2020-10-09 02:46:14.571696

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9ee998713b18"
down_revision = "3595c70244db"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_perth_year"
    )
    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_perth_month"
    )
    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_perth_day"
    )
    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_perth_hour"
    )

    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_sydney_year"
    )
    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_sydney_month"
    )
    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_sydney_day"
    )
    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_sydney_hour"
    )

    op.execute(
        "drop index if exists idx_balancing_summary_trading_interval_perth_year"
    )
    op.execute(
        "drop index if exists idx_balancing_summary_trading_interval_perth_month"
    )
    op.execute(
        "drop index if exists idx_balancing_summary_trading_interval_perth_day"
    )
    op.execute(
        "drop index if exists idx_balancing_summary_trading_interval_perth_hour"
    )


def downgrade():
    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_perth_year"
    )
    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_perth_month"
    )
    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_perth_day"
    )
    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_perth_hour"
    )

    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_sydney_year"
    )
    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_sydney_month"
    )
    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_sydney_day"
    )
    op.execute(
        "drop index if exists idx_facility_scada_trading_interval_sydney_hour"
    )

    op.execute(
        "drop index if exists idx_balancing_summary_trading_interval_perth_year"
    )
    op.execute(
        "drop index if exists idx_balancing_summary_trading_interval_perth_month"
    )
    op.execute(
        "drop index if exists idx_balancing_summary_trading_interval_perth_day"
    )
    op.execute(
        "drop index if exists idx_balancing_summary_trading_interval_perth_hour"
    )

    op.execute("COMMIT")

    op.create_index(
        "idx_facility_scada_trading_interval_perth_year",
        "facility_scada",
        [
            sa.text(
                "date_trunc('year', trading_interval AT TIME ZONE 'Australia/Perth')"
            )
        ],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_perth_month",
        "facility_scada",
        [
            sa.text(
                "date_trunc('month', trading_interval AT TIME ZONE 'Australia/Perth')"
            )
        ],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_perth_day",
        "facility_scada",
        [
            sa.text(
                "date_trunc('day', trading_interval AT TIME ZONE 'Australia/Perth')"
            )
        ],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_perth_hour",
        "facility_scada",
        [
            sa.text(
                "date_trunc('hour', trading_interval AT TIME ZONE 'Australia/Perth')"
            )
        ],
        postgresql_concurrently=True,
    )

    # NEM timezone indexes
    op.create_index(
        "idx_facility_scada_trading_interval_sydney_year",
        "facility_scada",
        [
            sa.text(
                "date_trunc('year', trading_interval AT TIME ZONE 'Australia/Sydney')"
            )
        ],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_sydney_month",
        "facility_scada",
        [
            sa.text(
                "date_trunc('month', trading_interval AT TIME ZONE 'Australia/Sydney')"
            )
        ],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_sydney_day",
        "facility_scada",
        [
            sa.text(
                "date_trunc('day', trading_interval AT TIME ZONE 'Australia/Sydney')"
            )
        ],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_sydney_hour",
        "facility_scada",
        [
            sa.text(
                "date_trunc('hour', trading_interval AT TIME ZONE 'Australia/Sydney')"
            )
        ],
        postgresql_concurrently=True,
    )



"""
    Queries for network data

    @TODO make these pluggable
    @TODO use sqlalchemy text() compiled queries
"""

from datetime import datetime
from typing import List

from opennem.core.networks import network_from_network_region
from opennem.core.normalizers import normalize_duid
from opennem.db.models.opennem import Network
from opennem.schema.network import NetworkSchema
from opennem.schema.time import TimeInterval, TimePeriod


def duid_in_case(facility_codes: List[str]) -> str:
    return ",".join(
        ["'{}'".format(i) for i in map(normalize_duid, facility_codes)]
    )


def power_facility(
    facility_codes: List[str],
    network_code: str,
    interval: TimeInterval,
    period: TimePeriod,
) -> str:

    scale = 1

    network = network_from_network_region(network_code)
    timezone = network.timezone_database

    if not timezone:
        timezone = "UTC"

    __query = """with intervals as (
            select generate_series(
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}') - '{period}'::interval,
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}'),
                '{interval}'::interval
            )::timestamp as interval
        )

        select
            i.interval as trading_day,
            fs.generated,
            fs.facility_code as facility_code
        from intervals i
        left outer join
            (select
                date_trunc('{trunc}', fs.trading_interval AT TIME ZONE '{timezone}')::timestamp  {interval_remainder} as interval,
                fs.facility_code,
                coalesce(avg(generated), 0) as generated
                from facility_scada fs
                where
                    fs.facility_code in ({facility_codes_parsed})
                    and fs.trading_interval > now() AT TIME ZONE '{timezone}' - '{period}'::interval
                    and fs.network_id = '{network_code}'
                group by 1, 2
            ) as fs on fs.interval = i.interval
        order by 1 desc, 2 desc"""

    query = __query.format(
        facility_codes_parsed=duid_in_case(facility_codes),
        network_code=network_code,
        trunc=interval.trunc,
        interval=interval.interval_human,
        interval_remainder=interval.get_sql_join(
            timezone=network.timezone_database
        ),
        period=period.period_sql,
        scale=scale,
        timezone=timezone,
    )

    return query


def power_network(
    network_code: str, interval: TimeInterval, period: TimePeriod,
) -> str:

    scale = 1

    network = network_from_network_region(network_code)
    timezone = network.timezone_database

    if not timezone:
        timezone = "UTC"

    __query = """with intervals as (
            select generate_series(
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}') - '{period}'::interval,
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}'),
                '{interval}'::interval
            )::timestamp as interval
        )

        select
            i.interval as trading_day,
            fs.generated,
            fs.facility_code as facility_code
        from intervals i
        left outer join
            (select
                date_trunc('{trunc}', fs.trading_interval AT TIME ZONE '{timezone}')::timestamp  {interval_remainder} as interval,
                fs.facility_code,
                coalesce(avg(generated), 0) as generated
                from facility_scada fs
                where
                    fs.trading_interval > now() AT TIME ZONE '{timezone}' - '{period}'::interval
                    and fs.network_id = '{network_code}'
                group by 1, 2
            ) as fs on fs.interval = i.interval
        order by 1 desc, 2 desc"""

    query = __query.format(
        network_code=network_code,
        trunc=interval.trunc,
        interval=interval.interval_human,
        interval_remainder=interval.get_sql_join(
            timezone=network.timezone_database
        ),
        period=period.period_sql,
        scale=scale,
        timezone=timezone,
    )

    return query


def power_network_fueltech(
    network: Network,
    interval: TimeInterval,
    period: TimePeriod,
    network_region: str = None,
) -> str:

    timezone = network.timezone_database

    if not timezone:
        timezone = "UTC"

    __query = """with intervals as (
            select generate_series(
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}') - '{period}'::interval,
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}'),
                '{interval}'::interval
            )::timestamp as interval
        )

        select
            i.interval as trading_day,
            fs.generated,
            fs.code
        from intervals i
        left outer join
            (select
                date_trunc('{trunc}', fs.trading_interval AT TIME ZONE '{timezone}')::timestamp  {interval_remainder} as interval,
                ft.code,
                coalesce(sum(generated), 0) as generated
                from facility_scada fs
                join facility f on fs.facility_code = f.code
                join fueltech ft on f.fueltech_id = ft.code
                where
                    fs.trading_interval > now() AT TIME ZONE '{timezone}' - '{period}'::interval
                    and fs.network_id = '{network_code}'
                    and f.fueltech_id is not null
                    {network_region_query}
                group by 1, 2
            ) as fs on fs.interval = i.interval
        order by 1 desc, 2 desc"""

    network_region_query = ""

    if network_region:
        network_region_query = f"and f.network_region='{network_region}'"

    query = __query.format(
        network_code=network.code,
        trunc=interval.trunc,
        interval=interval.interval_human,
        interval_remainder=interval.get_sql_join(
            timezone=network.timezone_database
        ),
        period=period.period_sql,
        network_region_query=network_region_query,
        timezone=timezone,
    )

    return query


def energy_facility(
    facility_codes: List[str],
    network_code: str,
    interval: TimeInterval,
    period: TimePeriod,
) -> str:

    network = network_from_network_region(network_code)
    timezone = network.timezone_database

    __query = """with intervals as (
            select generate_series(
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}') - '{period}'::interval,
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}'),
                '{interval}'::interval
            )::timestamp as interval
        )

        select
            i.interval as trading_day,
            fs.generated,
            fs.facility_code as facility_code
        from intervals i
        left outer join
            (select
                date_trunc('{trunc}', fs.trading_interval AT TIME ZONE '{timezone}')::timestamp  {interval_remainder} as interval,
                fs.facility_code,
                coalesce(sum(fs.eoi_quantity), NULL) / {scale} as generated
                from facility_scada fs
                where
                    fs.facility_code in ({facility_codes_parsed})
                    and fs.trading_interval > now() AT TIME ZONE '{timezone}' - '{period}'::interval
                    and fs.network_id = '{network_code}'
                group by 1, 2
            ) as fs on fs.interval = i.interval
        order by 1 desc, 2 asc"""

    query = __query.format(
        facility_codes_parsed=duid_in_case(facility_codes),
        network_code=network_code,
        trunc=interval.trunc,
        interval=interval.interval_sql,
        interval_remainder=interval.get_sql_join(
            timezone=network.timezone_database
        ),
        period=period.period_sql,
        scale=network.intervals_per_hour,
        timezone=timezone,
    )

    return query


def energy_network(
    network: NetworkSchema, interval: TimeInterval, period: TimePeriod,
) -> str:

    timezone = network.timezone_database

    __query = """with intervals as (
            select generate_series(
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}') - '{period}'::interval,
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}'),
                '{interval}'::interval
            )::timestamp as interval
        )

        select
            i.interval as trading_day,
            fs.generated,
            fs.facility_code as facility_code
        from intervals i
        left outer join
            (select
                date_trunc('{trunc}', fs.trading_interval AT TIME ZONE '{timezone}')::timestamp  {interval_remainder} as interval,
                fs.facility_code,
                coalesce(sum(fs.eoi_quantity), NULL) / {scale} as generated
                from facility_scada fs
                where
                    fs.trading_interval > now() AT TIME ZONE '{timezone}' - '{period}'::interval
                    and fs.network_id = '{network_code}'
                group by 1, 2
            ) as fs on fs.interval = i.interval
        order by 1 desc, 2 asc"""

    query = __query.format(
        network_code=network.code,
        trunc=interval.trunc,
        interval=interval.interval_sql,
        interval_remainder=interval.get_sql_join(
            timezone=network.timezone_database
        ),
        period=period.period_sql,
        scale=network.intervals_per_hour,
        timezone=timezone,
    )

    return query


def energy_network_fueltech(
    network: NetworkSchema,
    interval: TimeInterval,
    period: TimePeriod,
    network_region: str = None,
) -> str:

    timezone = network.timezone_database

    if not timezone:
        timezone = "UTC"

    __query = """with intervals as (
            select generate_series(
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}') - '{period}'::interval,
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}'),
                '{interval}'::interval
            )::timestamp as interval
        )

        select
            i.interval as trading_day,
            fs.generated,
            fs.code
        from intervals i
        left outer join
            (
                select
                    date_trunc('{trunc}', fs.trading_interval AT TIME ZONE '{timezone}')::timestamp  {interval_remainder} as interval,
                    ft.code,
                    coalesce(sum(fs.eoi_quantity), 0) / {scale} as generated
                from facility_scada fs
                join facility f on fs.facility_code = f.code
                join fueltech ft on f.fueltech_id = ft.code
                where
                    fs.trading_interval > now() AT TIME ZONE '{timezone}' - '{period}'::interval
                    and fs.network_id = '{network_code}'
                    and f.fueltech_id is not null
                    {network_region_query}
                group by 1, 2
            ) as fs on fs.interval = i.interval
        order by 1 desc, 2 desc"""

    network_region_query = ""

    if network_region:
        network_region_query = f"and f.network_region='{network_region}'"

    query = __query.format(
        network_code=network.code,
        trunc=interval.trunc,
        interval=interval.interval_sql,
        interval_remainder=interval.get_sql_join(
            timezone=network.timezone_database
        ),
        period=period.period_sql,
        scale=network.intervals_per_hour,
        network_region_query=network_region_query,
        timezone=timezone,
    )

    return query


def energy_network_fueltech_year(
    network: NetworkSchema,
    interval: TimeInterval,
    year: int,
    network_region: str = None,
) -> str:
    """
        copy + paste ftw .. for now - @TODO
    """

    timezone = network.get_timezone(postgres_format=True)

    if not timezone:
        timezone = "UTC"

    year_max = "{}-12-31".format(year)

    if year == datetime.now().year:
        year_max = "now()"

    __query = """
        SET SESSION TIME ZONE '{timezone}';

        select
            time_bucket_gapfill('{trunc}', trading_interval) AS trading_day,
            on_energy_sum(fs.generated) as energy,
            ft.code
        from facility_scada fs
        join facility f on fs.facility_code = f.code
        join fueltech ft on f.fueltech_id = ft.code
        where
            fs.trading_interval >= '{year}-01-01'
            and fs.trading_interval < {year_max}
            and fs.network_id = '{network_code}'
            and f.fueltech_id is not null
            {network_region_query}
        group by 1, 3
        order by 1 desc;
    """

    network_region_query = ""

    if network_region:
        network_region_query = f"and f.network_region='{network_region}'"

    query = __query.format(
        network_code=network.code,
        trunc=interval.interval_sql,
        year=year,
        year_max=year_max,
        scale=network.intervals_per_hour,
        network_region_query=network_region_query,
        timezone=timezone,
    )

    return query


def price_network_region(
    network: NetworkSchema,
    region_code: str,
    interval: TimeInterval,
    period: TimePeriod,
) -> str:

    timezone = network.timezone_database

    if not timezone:
        timezone = "UTC"

    __query = """with intervals as (
            select generate_series(
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}') - '{period}'::interval,
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}'),
                '{interval}'::interval
            )::timestamp as interval
        )

        select
            i.interval as trading_day,
            fs.generated,
            fs.network_region
        from intervals i
        left outer join
            (select
                date_trunc('{trunc}', fs.trading_interval AT TIME ZONE '{timezone}')::timestamp  {interval_remainder} as interval,
                fs.network_region,
                coalesce(avg(fs.price), NULL) as generated
                from balancing_summary fs
                where
                    fs.trading_interval > now() AT TIME ZONE '{timezone}' - '{period}'::interval
                    and fs.network_id = '{network_code}'
                    and fs.network_region = '{network_region}'
                group by 1, 2
            ) as fs on fs.interval = i.interval
        order by 1 desc, 2 asc"""

    query = __query.format(
        network_code=network.code,
        network_region=region_code,
        trunc=interval.trunc,
        interval=interval.interval_sql,
        interval_remainder=interval.get_sql_join(
            timezone=network.timezone_database
        ),
        period=period.period_sql,
        scale=network.intervals_per_hour,
        timezone=timezone,
    )

    return query



import logging
from datetime import datetime
from pprint import pprint

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import text

from opennem.db import db_connect
from opennem.db.load_fixtures import update_existing_geos
from opennem.db.models.opennem import Facility, Station
from opennem.geo.google_geo import place_search

logger = logging.getLogger(__name__)


def build_address_string(station_record):
    l = [
        station_record.network_name,
        station_record.locality,
        station_record.state,
        "Australia",
    ]

    address_string = ", ".join(str(i) for i in l if i)

    return address_string


def opennem_geocode(limit=None):
    engine = db_connect()
    session = sessionmaker(bind=engine)
    s = session()

    records = (
        s.query(Station)
        .filter(Station.geom == None)
        .filter(Station.geocode_skip == False)
    )

    count = 0
    skipped = 0
    records_added = 0

    for r in records:
        geo_address_string = build_address_string(r)

        logger.info("Geocoding record: {}".format(geo_address_string))
        continue

        google_result = place_search(geo_address_string)

        pprint(google_result)

        if (
            google_result
            and type(google_result) is list
            and len(google_result) > 0
        ):
            result = google_result[0]

            r.place_id = result["place_id"]

            lat = result["geometry"]["location"]["lat"]
            lng = result["geometry"]["location"]["lng"]
            r.geom = "SRID=4326;POINT({} {})".format(lng, lat)

            r.geocode_processed_at = datetime.now()
            r.geocode_by = "google"
            r.geocode_approved = False

            try:
                s.add(r)
                s.commit()
                records_added += 1
            except IntegrityError as e:
                logger.error(e)
                skipped += 1
                pass
            except Exception as e:
                skipped += 1
                logger.error("Error: {}".format(e))
        else:
            skipped += 1

        count += 1
        if limit and count >= limit:
            break

    print(
        "Geocode of opennem records done. Added {} records. Couldn't match {}".format(
            records_added, skipped
        )
    )


if __name__ == "__main__":
    update_existing_geos()
    opennem_geocode()



"""Create hypertables

Revision ID: ad3938cfbd45
Revises: 28a206c84987
Create Date: 2020-10-16 22:28:23.901036

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "ad3938cfbd45"
down_revision = "28a206c84987"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        "select create_hypertable('facility_scada', 'trading_interval', if_not_exists => TRUE, migrate_data => TRUE, chunk_time_interval => INTERVAL '1 week')"
    )
    op.execute(
        "select create_hypertable('balancing_summary', 'trading_interval', if_not_exists => TRUE, migrate_data => TRUE, chunk_time_interval => INTERVAL '1 week')"
    )


def downgrade():
    pass



from datetime import datetime, timedelta

import scrapy

from opennem.pipelines.apvi.data import APVIStoreData
from opennem.utils.dates import date_series, get_date_component

APVI_DATA_URI = "https://pv-map.apvi.org.au/data"

TODAY = datetime.now().date()

YESTERDAY = TODAY - timedelta(days=1)

MONTH_AGO = TODAY - timedelta(days=29)

APVI_DATE_QUERY_FORMAT = "%Y-%m-%d"

APVI_EARLIEST_DATE = "2013-05-07"


class APVIDataSpiderBase(scrapy.Spider):

    name = "au.apvi.latest.data"

    pipelines = set([APVIStoreData])

    def start_requests(self):
        yield scrapy.FormRequest(
            APVI_DATA_URI,
            formdata={"day": get_date_component(APVI_DATE_QUERY_FORMAT)},
            meta={
                "is_latest": True,
                "record_date": get_date_component(APVI_DATE_QUERY_FORMAT),
            },
        )

    def parse(self, response):
        yield {"records": response.json(), "meta": response.meta}


class APVIDataCurrentSpider(APVIDataSpiderBase):
    name = "au.apvi.current.data"

    def start_requests(self):
        for date in date_series(TODAY, length=31, reverse=True):
            yield scrapy.FormRequest(
                APVI_DATA_URI,
                formdata={"day": date.strftime(APVI_DATE_QUERY_FORMAT)},
                meta={"is_latest": True if date == TODAY else False},
            )


class APVIDataArchiveSpider(APVIDataSpiderBase):
    name = "au.apvi.archive.data"

    def start_requests(self):
        end_date = datetime.strptime(
            APVI_EARLIEST_DATE, APVI_DATE_QUERY_FORMAT
        ).date()

        for date in date_series(start=MONTH_AGO, end=end_date, reverse=True):
            yield scrapy.FormRequest(
                APVI_DATA_URI,
                formdata={"day": date.strftime(APVI_DATE_QUERY_FORMAT)},
            )

    custom_settings = {
        "CONCURRENT_REQUESTS": 1,
        "DOWNLOAD_DELAY": 1,
    }



from opennem.pipelines.nem.opennem import NemwebUnitScadaOpenNEMStorePipeline
from opennem.spiders.nemweb import NemwebSpider


class NemwebCurrentDispatch(NemwebSpider):
    name = "au.nem.current.dispatch"
    start_url = "http://nemweb.com.au/Reports/Current/Next_Day_Dispatch/"

    # only get most recent
    limit = 1

    pipelines_extra = set([NemwebUnitScadaOpenNEMStorePipeline,])


class NemwebArchiveDispatch(NemwebSpider):
    name = "au.nem.archive.dispatch"
    start_url = "http://nemweb.com.au/Reports/Current/Next_Day_Dispatch/"
    limit = 0
    skip = 1

    pipelines_extra = set([NemwebUnitScadaOpenNEMStorePipeline,])

    # Archives tend to contain large zips of embedded zips so throttle
    # to limit memory use
    custom_settings = {
        "CONCURRENT_REQUESTS": 4,
        "CONCURRENT_ITEMS": 1,
    }



import datetime
import pandas as pd
from dms_daemon import orm
from dms_daemon.nem_derived import meta


#supress data truncation warnings
import MySQLdb
import warnings
warnings.filterwarnings('ignore', category=MySQLdb.Warning)

NEMWEB_LIVE = orm.engine2
NEMWEB_META = orm.engine_meta
NEMWEB_DERIVED = orm.engine_derived

def select_price(d=datetime.date(2018,6,1)):
    sql =     "SELECT SETTLEMENTDATE, RRP, REGIONID "\
            "FROM `TRADING_PRICE` "\
            "WHERE SETTLEMENTDATE > '{0}' "\
                "AND SETTLEMENTDATE <= '{1}'"
    return pd.read_sql(    sql.format(d,d+datetime.timedelta(1)), 
                        con=NEMWEB_LIVE)#, index_col="SETTLEMENTDATE")
                        
def select_regionsum(d=datetime.date(2018,6,1)):
    sql =     "SELECT SETTLEMENTDATE, TOTALDEMAND, NETINTERCHANGE, DEMAND_AND_NONSCHEDGEN, REGIONID "\
            "FROM `TRADING_REGIONSUM` "\
            "WHERE SETTLEMENTDATE > '{0}' "\
                "AND SETTLEMENTDATE <= '{1}'"
    return pd.read_sql(    sql.format(d,d+datetime.timedelta(1)), 
                        con=NEMWEB_LIVE)
                        
def select_output(d=datetime.date(2018,6,1)):
    sql =     "SELECT SETTLEMENTDATE, OUTPUT_MWH, QUALITY_FLAG, DUID "\
            "FROM `TRADING_DUID_OUTPUT` "\
            "WHERE SETTLEMENTDATE > '{0}' "\
                "AND SETTLEMENTDATE <= '{1}'"
    return pd.read_sql(    sql.format(d,d+datetime.timedelta(1)), 
                        con=NEMWEB_DERIVED)#, index_col="SETTLEMENTDATE")
                        
def select_rooftop(d=datetime.date(2018,6,1)):
    sql =     "SELECT INTERVAL_DATETIME as SETTLEMENTDATE, REGIONID, POWER as ROOFTOP_SOLAR "\
            "FROM `ROOFTOP_ACTUAL` "\
            "WHERE INTERVAL_DATETIME > '{0}' "\
                "AND INTERVAL_DATETIME <= '{1}'"
    return pd.read_sql(    sql.format(d,d+datetime.timedelta(1)), 
                        con=NEMWEB_LIVE)
def category_map():
    sql =     "SELECT NAME,ID FROM CATEGORY"
    results = NEMWEB_DERIVED.execute(sql)
    return dict(results.fetchall())
    
def mapper(i):    
    duid = meta.DUID_KEYS[i]
    fcas = {'APD01': 5, 'SNOWYGJP': 5, 'ASNENC1':1, 'VENUS1':5, 'ASQENC1':2, 'ASTHYD1': 4}
    if i in meta.REGION_MAP:
        return meta.REGION_MAP[i]
    elif "_" in duid:    
        d = {"NS":1, "QL":2, "SA":3, "TA":4, "VI":5}
        return d[duid.split("_")[1][:2]]
    elif duid in fcas: 
        return fcas[duid]
    else:
        raise KeyError (duid,i)
        
def select_day(d=datetime.date(2018,6,1)):
    df = select_output(d=d)    
    df['REGIONID'] = df.DUID.apply(lambda x: mapper(x))
    df_p = select_price(d=d)
    return pd.merge(df,df_p, left_on=["SETTLEMENTDATE", "REGIONID"], right_on=["SETTLEMENTDATE","REGIONID"])

def revenue_generator(d):
    df = select_day(d)
    for duid,g in df.groupby("DUID"):
       yield (d,duid, g.OUTPUT_MWH.sum(), (g.OUTPUT_MWH*g.RRP).sum(), g.QUALITY_FLAG.max())

def calculate_day(d=datetime.date(2018,6,1)):
    return pd.DataFrame([i for i in revenue_generator(d)],columns=['DATE','DUID', 'OUTPUT_MWH', 'MARKET_VALUE', "QUALITY_FLAG"])

def insert_daily_duid(d):    
    df = calculate_day(d)
    df.to_sql(name="DAILY_DUID_OUTPUT", index=None, con=NEMWEB_DERIVED, if_exists='append')
    
def select_region_day(d=datetime.date(2018,6,1)):
    df = select_regionsum(d=d)    
    df_p = select_price(d=d)
    df_r = select_rooftop(d=d)
    df_tmp = pd.merge(df,df_p, left_on=["SETTLEMENTDATE", "REGIONID"], right_on=["SETTLEMENTDATE","REGIONID"])
    return     pd.merge(df_tmp, df_r, left_on=["SETTLEMENTDATE", "REGIONID"], right_on=["SETTLEMENTDATE","REGIONID"])
    
def split_interchange(df):
    df['NETINTERCHANGE_IMPORT'] = df.NETINTERCHANGE[df.NETINTERCHANGE<=0]
    df['NETINTERCHANGE_EXPORT'] =  df.NETINTERCHANGE[df.NETINTERCHANGE>=0]
    df.fillna(0,inplace=True)
    
def region_summary(df, date, regionid):
    columns = list(meta.CATEGORY_MAP)
    
    energy = df[columns].sum()/2
    energy.name = "ENERGY_MWH"
    
    market_value = df[columns].multiply(df.RRP, axis='index').sum()/2
    market_value.name = "MARKET_VALUE"
    
    dx = pd.DataFrame([energy,market_value]).transpose()
    dx.index.name = 'CATEGORY_ID'    
    dx.reset_index(inplace=True)
    
    dx.CATEGORY_ID = dx.CATEGORY_ID.apply(lambda x: meta.CATEGORY_MAP[x])
    dx['DATE'] = date
    dx['REGIONID'] = regionid
    
    return dx
    
def insert_daily_region_summary(d):
    df = select_region_day(d = d)
    for regionid, df_region in df.groupby(df.REGIONID):
        df_region = df_region.copy()
        split_interchange(df_region)
        dx = region_summary(df_region, d, regionid)
        dx.to_sql(name="DAILY_REGION_SUMMARY", index=None, con=NEMWEB_DERIVED, if_exists='append')
    #working version, completed from 1st May 2015

def daily_aggregator():
    todays_date = datetime.date.today()
    aggregate_duid(todays_date)
    aggregate_regionsum(todays_date)

def aggregate_duid(todays_date):
    sql = "SELECT MAX(DATE) FROM DAILY_DUID_OUTPUT"
    latest_date = NEMWEB_DERIVED.execute(sql).fetchall()[0][0]
    next_date = latest_date + datetime.timedelta(1)
    while next_date < todays_date:
        insert_daily_duid(next_date)
        next_date += datetime.timedelta(1)

def aggregate_regionsum(todays_date):
    sql = "SELECT MAX(DATE) FROM DAILY_REGION_SUMMARY"
    latest_date = NEMWEB_DERIVED.execute(sql).fetchall()[0][0]
    next_date = latest_date + datetime.timedelta(1)
    while next_date < todays_date:
        insert_daily_region_summary(next_date)
        next_date += datetime.timedelta(1)



import io
import logging
import os

logger = logging.getLogger(__name__)

LOCAL_DIR_PATH = os.path.join(os.path.dirname(__file__), "..", "..", "data")

if not os.path.exists(LOCAL_DIR_PATH):
    raise Exception("Folder {} does not exist".format(LOCAL_DIR_PATH))


def write_to_local(file_path, data):
    save_file_path = os.path.join(LOCAL_DIR_PATH, file_path)

    if type(data) is io.StringIO:
        data = data.getvalue()

    with open(save_file_path, "w") as fh:
        fh.write(data)

    logger.info("Wrote {} to {}".format(len(data), save_file_path))



import datetime
import pandas as pd
from sqlalchemy import create_engine
from dms_daemon.nem_derived import meta
from dms_daemon import orm

#supress data truncation warnings
import MySQLdb
import warnings
warnings.filterwarnings('ignore', category=MySQLdb.Warning)

engine = orm.engine_derived


def duid_dataframe_generator(df,columns = ['DUID','INITIALMW','QUALITY_FLAG']):
    for duid, dx in df.groupby('DUID'):
        yield DUID(dx[columns],columns[1])
        
def nsduid_dataframe_generator(df,columns = ['DUID','MWH_READING','QUALITY_FLAG']):
    for duid, dx in  df.groupby('DUID'):
        yield NonSchedDUID(dx[columns],columns[1])
        
def data_check(df):
    for ds in duid_dataframe_generator(df):
        try:
            ds.validate_scada_data()
        except Exception as E:
            #first full day of new duid (with no midnight recording)
            if (ds.index[0].time() == datetime.time(0,5)) & (ds.data_length==289):
                continue
            #day before first full day (of new duid)
            elif (ds.index[0].time() == datetime.time(0,5)) & (ds.data_length==1):
                continue
            #last full day of old duid (with no 5 past midnight recording)            
            elif (ds.index[-1].time() == datetime.time(0,0)) & (ds.data_length==289):
                continue
            #day after last full day of old new day 
            elif (ds.index[-1].time() == datetime.time(0,0)) & (ds.data_length==1):                
                continue
            #nothing before hand                
            elif (ds.duid[:3] == "RT_") & (ds.sum()==0):
                #might be data here
                continue
            elif check_duid_range(ds, check="MIN") == True:
                #might be data here
                continue                
            #nothing afterwards                 
            elif check_duid_range(ds, check="MAX") == True:
                #might be data here
                continue
            else:
                raise

def get_solution(d1=datetime.datetime(2017,1,1),db='nemweb'):
    d2 = d1 + datetime.timedelta(1,300)
    sql =    "SELECT S.SETTLEMENTDATE, S.DUID, S.INITIALMW, S.INTERVENTION, 0 as QUALITY_FLAG "\
            "FROM DISPATCH_UNIT_SOLUTION S "\
            "INNER JOIN ("\
                "SELECT SETTLEMENTDATE, DUID, MAX(INTERVENTION) as I_FLAG "\
                "FROM DISPATCH_UNIT_SOLUTION "\
                "WHERE SETTLEMENTDATE >= '{0}' "\
                    "AND SETTLEMENTDATE <=  '{1}' "\
                "GROUP BY SETTLEMENTDATE, DUID) AS S2 "\
            "ON S2.SETTLEMENTDATE = S.SETTLEMENTDATE "\
                "AND S2.DUID = S.DUID "\
                "AND S2.I_FLAG = S.INTERVENTION".format(d1,d2)
    df = pd.read_sql(sql, con=orm.engine2)
    return DailyDispatch(d1,df)    
    
def get_non_schedgen(d1=datetime.datetime(2017,1,1),db='nemweb'):
    d2 = d1 + datetime.timedelta(1,300)
    sql =    "SELECT INTERVAL_DATETIME AS SETTLEMENTDATE, DUID, MWH_READING, 0 as QUALITY_FLAG "\
            "FROM METER_DATA_GEN_DUID "\
            "WHERE INTERVAL_DATETIME >= '{0}' "\
                "AND INTERVAL_DATETIME <=  '{1}'".format(d1,d2)
    df = pd.read_sql(sql, con=orm.engine2)
    return df
    #return DailyDispatch(d1,df)    
                
class DailyDispatch(pd.DataFrame):
    """ pandas DataFrame sublcass 
        - additional meta data (date of dispatch data)
        - additional methods to process (insert) duid data"""
    def __init__(self, d1, df):
        pd.DataFrame.__init__(self,df)                
        self.date = d1.date()
        self.d1 = d1
        self.d2 = self.d1 + datetime.timedelta(1,300)
        
    def data_transaction(self):    
        connection =  engine.connect()
        self.__transaction =  connection.begin()
        
        try:
            self.process_data()
            self.__transaction.commit()                    
            connection.close()
        except Exception as E:
            self.__transaction.rollback()
            connection.close()
            raise E
            
    def process_data(self):
        for duid in duid_dataframe_generator(self):
            if (duid.data_length == 1) & (duid.df.INITIALMW.sum()==0):
                continue
            if (duid.data_length == 1) & (duid.df.index[0]==datetime.datetime(2011,5,17)) & (duid.duid_id == 63):        
                continue
            if (duid.data_length == 1) & (duid.df.index[0]==datetime.datetime(2014,8,29,0,5)) & (duid.duid_id == 293):        
                continue

            if (duid.data_length == 1) & (duid.df.index[0]==datetime.datetime(2015,7,1,0,5)) & (duid.duid_id == 339):        
                continue

            if (duid.data_length == 1) & (duid.df.index[0]==datetime.datetime(2015,7,1,0,5)) & (duid.duid_id == 349):        
                continue

            if (duid.data_length == 1) & (duid.df.index[0]==datetime.datetime(2015,7,1,0,5)) & (duid.duid_id == 350):        
                continue

            if (duid.data_length == 1) & (duid.df.index[0]==datetime.datetime(2015,7,1)) & (duid.duid_id == 218):        
                continue

            if (duid.data_length == 1) & (duid.df.index[0]==datetime.datetime(2015,7,1)) & (duid.duid_id == 236):        
                continue

            if (duid.data_length == 1) & (duid.df.index[0]==datetime.datetime(2010,8,20,0,5)) & (duid.duid_id == 161):
                #waterloo wf        
                continue            
                
            if (duid.data_length == 1) & (duid.df.index[0]==datetime.datetime(2009,9,22)) & (duid.duid_id == 298):
                #mckay2 gt 
                continue                        
            
            if (duid.data_length == 1) & (duid.df.index[0]==datetime.datetime(2019,6,19,0,5)) & (duid.duid_id == 749):
                #first_day lal lal / Yendoon
                continue                        

            if (duid.data_length == 1) & (duid.df.index[0]==datetime.datetime(2020,4,7,0,5)) & (duid.duid_id == 769):
                #Elaine
                continue
                
            if (duid.df.index[0]==datetime.datetime(2019,10,22)) & (duid.duid_id == 3):
                #Blowering problem (missing day)
                continue                        

            if (duid.df.index[0]==datetime.datetime(2019,10,23,4,5)) & (duid.duid_id == 3):
                #Blowering problem (missing day = 22-23)
                continue                        

            else: 
                duid.validate_scada_data(self.d1,self.d2)                
                self.__insert_duid_data__(duid)
                    
    def __insert_duid_data__(self,duid,test=""):
        df_t =duid.trading_energy_data()        
        df_t.to_sql(name="TRADING_DUID_OUTPUT{}".format(test), index=None, con=self.__transaction.connection, if_exists='append')        
        
class DailyNonSched(DailyDispatch): 
    def __init__(self, d1, df):
        DailyDispatch.__init__(self,d1,df)    
        
    def process_data(self):
        for duid in nsduid_dataframe_generator(self, columns=['DUID','MWH_READING','QUALITY_FLAG']):        
            if (duid.duid == 'GERMCRK') & (duid.df.index[0].date() == datetime.date(2018,4,9)):
                df = german_creek(duid.df, self.d1, self.d2, backward=False)
                gc_interpolater_one(df, test="")#_TEST        
                continue
            try:            
                duid.validate_scada_data(self.d1,self.d2)
            except:                    
                try:
                    duid._NonSchedDUID__interpolate_missing_data(self.d1,self.d2)
                except:
                    not_na = duid.df[duid.df.MWH_READING.notna()]
                    if (duid.data_length == 1) & (not_na.index[0].time() == datetime.time(0,5)):
                        continue
                    elif (duid.duid == "LRSF1") & (duid.df.index[0].date() <= datetime.date(2018,5,5)):
                        print("LRSF1 commissioning")
                        continue
                    elif (duid.duid == 'GERMCRK') & (duid.df.index[0].date() == datetime.date(2018,4,12)):            
                        df = german_creek(duid.df, self.d1, self.d2, backward=True)
                        gc_interpolater(df, test="")#_TEST        
                        continue                    
                    else:
                        raise Exception ("New issue", duid, duid.duid_id, duid.df.index[0].date())
            self.__insert_duid_data__(duid,test="")#_TEST        
            
class DUID():
    """Class for storing a days worth of duid data"""
    def __init__(self,dataframe,data_col='INITIALMW'):            
        self.duid_id = dataframe.DUID[0]
        self.duid = meta.DUID_KEYS[self.duid_id]
        self.df = dataframe    
        self.data_length = len(set(self.df.index))
        self.data_col = data_col
    
    def __repr__(self):
        return "{0}".format(self.duid)
        
    def validate_scada_data(self,d1,d2):                

        if self.data_length < 290:
            self.__process_missing_data(d1,d2)
            
        elif self.data_length > 290:
            if self.df.index[0] == datetime.datetime(2001,4,8):
                #huge gap over 2 days (2001-04-08)                
                self.__interpolate_missing_data(d1,d2,quality_flag=2)
            else:        
                raise ExtraDataError(self.data_length,self.duid_id,self.df.index[0].date())                    
            
            
    def __process_missing_data(self,d1,d2):
        #last data point missing (e.g. where duid retired)
        #need to interpolate (fill) last value (quality unaffected, flag is zero)
        if (self.df.index[-1].time() == datetime.time(0,0)) & (self.data_length==289):                
            self.__interpolate_missing_data(d1,d2,quality_flag=0)            
        
        #first day, new duid (duid = 639, HDWF3)
        elif (self.df[self.data_col].sum()==0) & (self.duid_id == 639):                
            self.__interpolate_missing_data(d1,d2,quality_flag=0)
            
        #first day, new duid (duid = 691)
        elif (self.df.index[0]==datetime.datetime(2018,7,19,0,5)) & (self.duid_id == 691):                
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #first day, new duid (duid = 133, LKBONNY2)
        elif (self.df.index[0]==datetime.datetime(2007,5,31,10,55)) & (self.duid_id == 133):                
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #first day, new duid (duid = 86, KPP_1)
        elif (self.df.index[0]==datetime.datetime(2007,4,3,10,50)) & (self.duid_id == 86):                
            self.__interpolate_missing_data(d1,d2,quality_flag=0)    
            
        #first day, new duid (duid = 164,165,166, BBTHREE 1,2,3)
        elif (self.df.index[0]==datetime.datetime(2006,6,8,11,0)) & (self.duid_id >= 164) & (self.duid_id <= 166):                
            self.__interpolate_missing_data(d1,d2,quality_flag=0)
            
        #first day, new duid (duid = 56 Braemar1)
        elif (self.df.index[0]==datetime.datetime(2006,4,26,10,15)) & (self.duid_id >= 56) & (self.duid_id <= 58):                
            self.__interpolate_missing_data(d1,d2,quality_flag=0)
            
        #first day, new duid (duid = 218 LAVNORTH)
        elif (self.df.index[0]==datetime.datetime(2006,3,2,16,20)) & (self.duid_id == 218):                
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #first day, new duid (duid = 236 VPGS)
        elif (self.df.index[0]==datetime.datetime(2006,1,31,13,35)) & (self.duid_id == 236):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #last day, old duids (duid = 336, 337, 338, 339, 340, 341)
        elif (self.df.index[0]==datetime.datetime(2006,1,31,0,0)) & (self.duid_id >= 336) & (self.duid_id <= 341):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #skip RTs, where sum = zero
        elif (self.duid[:3] == "RT_") & (self.df[self.data_col].sum()==0):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #skip DGs, where sum = zero (dummy generators)
        elif (self.duid[:3] == "DG_") & (self.df[self.data_col].sum()==0):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #first day (bastyon, 163)         #possible re-registration as scheduled?
        elif (self.df.index[0]==datetime.datetime(2005,5,16,13,35)) & (self.duid_id == 163):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #first day (CETHUNA, 168) possible re-registration as scheduled?
        elif (self.df.index[0]==datetime.datetime(2005,5,16,13,35)) & (self.duid_id == 168):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #first day (DEVILS_G, 170) possible re-registration as scheduled - whole heap of tas hydro generator re-classified)
        elif (self.df.index[0]==datetime.datetime(2005,5,16,13,35)) & (self.duid_id >= 170) & (self.duid_id <= 190):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #first day (ANGAS1 & ANGAS2, 119,120)
        elif (self.df.index[0]==datetime.datetime(2004,12,30,12,55)) & (self.duid_id >= 119) & (self.duid_id <= 120):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #first day (YABULU2, 116)
        elif (self.df.index[0]==datetime.datetime(2004,10,21,11,5)) & (self.duid_id == 116):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #first day (BellBay3, 342) possible re-registration as scheduled - whole heap of tas hydro generator re-classified)
        elif (self.df.index[0]==datetime.datetime(2005,5,16,13,35)) & (self.duid_id >= 342) & (self.duid_id <= 343):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)


        #last day (swanbank D, 314) 
        elif (self.df.index[0]==datetime.datetime(2004,6,1)) & (self.duid_id == 314):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #last day (swanbank E, 313)
        elif (self.df.index[0]==datetime.datetime(2003,3,4)) & (self.duid_id == 313):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #last day (TNPS1, 112)
        elif (self.df.index[0]==datetime.datetime(2002,9,16,17,5)) & (self.duid_id == 112):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #last day ('MRIDGE', 299)
        elif (self.df.index[0]==datetime.datetime(2002,7,16,0,0)) & (self.duid_id == 299):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #first day (Swanbank E, 107)
        elif (self.df.index[0]==datetime.datetime(2002,5,15,10,35)) & (self.duid_id == 107):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #first day (AGLSOM, 193)
        elif (self.df.index[0]==datetime.datetime(2002,1,10,10,5)) & (self.duid_id == 193):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #first day (AGLHAL, 118)
        elif (self.df.index[0]==datetime.datetime(2001,12,18,9,5)) & (self.duid_id == 118):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #first day (BDL02, 196)
        elif (self.df.index[0]==datetime.datetime(2001,12,13,8,5)) & (self.duid_id == 196):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #first day (LYGS1, 330)
        elif (self.df.index[0]==datetime.datetime(2001,12,12,8,5)) & (self.duid_id >= 330) & (self.duid_id <= 335):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)


        #last day SWAN A, 307)
        elif (self.df.index[0]==datetime.datetime(2002,7,16,0,0)) & (self.duid_id >= 307) & (self.duid_id <= 313):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)
            
            
        #first day QPS1, 146
        elif (self.df.index[0]==datetime.datetime(2001,11,8,10,50)) & (self.duid_id >= 145) & (self.duid_id <= 148):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)
            
        #last day NGTS, 300
        elif (self.df.index[0]==datetime.datetime(2001,10,9)) & (self.duid_id == 300):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)
            
        #last day SNOWY 1-6, 301-306
        elif (self.df.index[0]==datetime.datetime(2001,9,24)) & (self.duid_id >= 301) & (self.duid_id <= 306):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)


        #first data point missing (e.g. where duid first added)    
        #can ignore first data point (not used in interpolation)
        elif (self.df.index[0].time() == datetime.time(0,5)) & (self.data_length==289):                        
            pass
            
            
        #actual gap in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2001,9,25)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)


        #actual gap in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2001,8,20)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)            
        
        #actual gap in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2001,8,7)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
        
        #actual gap in the data!     ("known gap")    
        elif (self.df.index[0]==datetime.datetime(2001,6,18)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
            
        #actual gap in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2001,8,6)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
            
        #actual gap in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2001,7,5)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
            
        #two gaps in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2001,6,5)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
            
        #three gaps in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2001,5,21)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)            
            
        #two gaps in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2001,3,28)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)            
            
        #three gaps in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2001,2,27)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)        
            
        #single gaps in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2001,2,19)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)        
            
        #double gap in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2001,2,5)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)        
            
        #two gaps in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2000,12,7)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)        
    
        #four gaps in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2000,8,24)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)        

        #big gap in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2000,8,23)):
            self.__interpolate_missing_data(d1,d2,quality_flag=2)        

        # gap in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2000,8,22)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
            
        # gap in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2000,8,1)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)

        # gap in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2000,7,17)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)    
            
        # big gap in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2000,5,13)):
            self.__interpolate_missing_data(d1,d2,quality_flag=2)    

        # gap in the data!     ("known gap")
        elif (self.df.index[0]==datetime.datetime(2000,6,27)):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)    
            
        # BW01 out of service or similar?
        elif (self.duid_id ==5 ) & (self.df.index[0]==datetime.datetime(2000,5,9,12)):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)
            
        #HUME NSW
        elif (self.duid_id ==22 ) & (self.df.index[0]==datetime.datetime(2000,5,9,13,35)):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)
            
        #HVGTS NSW
        elif (self.duid_id ==23 ) & (self.df.index[0]==datetime.datetime(2000,5,9,13,35)):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)
            
        #LD01
        elif (self.duid_id ==25 ) & (self.df.index[0]==datetime.datetime(2000,5,9,13,35)):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)

        #MM4
        elif (self.duid_id ==30 ) & (self.df.index[0]==datetime.datetime(2000,5,9,13,35)):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)
        
        #??
        elif (self.df[self.data_col].sum() == 0 ) & (self.df.index[0].date() == datetime.date(2000,5,9)):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)
            
        #pump1
        elif (self.duid_id==49) & (self.df.index[0]==datetime.datetime(2000,5,9,3,35)):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)
            
        #BARCALDN
        elif (self.duid_id==53) & (self.df.index[0]==datetime.datetime(2000,5,9,6,5)):
            self.__interpolate_missing_data(d1,d2,quality_flag=0)                
                
        #first non_sched_gen
        elif (self.duid_id==277) & (self.df.index[0]==datetime.datetime(2016,1,1,0,0)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            #gap of 1.5 hours
            self.__interpolate_missing_data(d1,d2,quality_flag=2)                    
            
        #ANGAS non_sched_gen
        elif (self.duid_id>=119) & (self.duid_id<=120) & (self.df.index[0]==datetime.datetime(2016,1,5,0,0)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            #long gap, but zero output
            self.__interpolate_missing_data(d1,d2,quality_flag=2)            
            
        #CHALLHWF non_sched_gen
        elif (self.duid_id==197) & (self.df.index[0]==datetime.datetime(2016,1,12,0,0)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            #long gap, but zero output
            self.__interpolate_missing_data(d1,d2,quality_flag=2)            
            
        #Angas non-sched
        elif (self.duid_id>=119) & (self.duid_id<=120) & (self.df.index[0]==datetime.datetime(2016,1,15,0,0)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            #long gap, but zero output
            self.__interpolate_missing_data(d1,d2,quality_flag=1)            
        
        #Angas non-sched
        elif (self.duid_id>=119) & (self.duid_id<=120) & (self.df.index[0]==datetime.datetime(2016,1,16,1,25)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            #long gap, but zero output
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
            
        #CLOVER non_sched_gen
        elif (self.duid_id==198) & (self.df.index[0]==datetime.datetime(2016,1,17,0,0)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            #long gap, but zero output
            self.__interpolate_missing_data(d1,d2,quality_flag=1)            
            
        #PORTWF non_sched_gen
        elif (self.duid_id==234) & (self.df.index[0]==datetime.datetime(2016,1,17,0,0)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            #long gap, but zero output
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
            
        #WAUBRAWF non_sched_gen
        elif (self.duid_id==237) & (self.df.index[0]==datetime.datetime(2016,1,17,0,0)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            #long gap, but zero output
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
            
        #MLWF non_sched_gen
        elif (self.duid_id==277) & (self.df.index[0]==datetime.datetime(2016,1,17,0,0)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
            
        #GERMCRK non_sched_gen
        elif (self.duid_id==280) & (self.df.index[0]==datetime.datetime(2016,1,22,0,0)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
            
        #GERMCRK non_sched_gen
        elif (self.duid_id==280) & (self.df.index[0]==datetime.datetime(2016,1,23,0,25)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            #day long gap
            self.__interpolate_missing_data(d1,d2,quality_flag=3)
            
        #GERMCRK non_sched_gen
        elif (self.duid_id==280) & (self.df.index[0]==datetime.datetime(2016,1,25,9,5)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            #day long gap
            self.__interpolate_missing_data(d1,d2,quality_flag=3)
            
        #YAMBUKWF non_sched_gen
        elif (self.duid_id==240) & (self.df.index[0]==datetime.datetime(2016,1,31)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            self.__interpolate_missing_data(d1,d2,quality_flag=2)
            
        #YAMBUKWF non_sched_gen
        elif (self.duid_id==240) & (self.df.index[0]==datetime.datetime(2016,2,6)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
            
        #YAMBUKWF non_sched_gen
        elif (self.duid_id==240) & (self.df.index[0]==datetime.datetime(2016,2,7,6,10)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            self.__interpolate_missing_data(d1,d2,quality_flag=2)
            
        #MLWF non_sched_gen
        elif (self.duid_id==277) & (self.df.index[0]==datetime.datetime(2016,2,7)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
            
        #ANGAS non_sched_gen
        elif (self.duid_id>=119) & (self.duid_id<=120) & (self.df.index[0]==datetime.datetime(2016,2,10)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            #zero)
            self.__interpolate_missing_data(d1,d2,quality_flag=0)
            
        #Generic ANGAS
        elif (self.duid_id>=119) & (self.duid_id<=120):
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
            print ("Zero interpolation for ANGAS")
            if self.df[self.df.QUALITY_FLAG==1].max()[self.data_col] != 0:                
                raise MissingDataError(self.data_length,self.duid_id,self.df.index[0].date())
                
        #YAMBUKWF non_sched_gen
        elif (self.duid_id==240) & (self.df.index[0]==datetime.datetime(2016,2,13,0,5)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
            
            
        elif (self.df.index[0].time() ==datetime.time(0,0)) & (self.df.index[-1].time() ==datetime.time(0,5)):
            if self.data_length>285:
                self.__interpolate_missing_data(d1,d2,quality_flag=1)    
                print ("Interpolated data:", self.duid_id, self.duid, self.df.index[0].date())
            else:
                raise MissingDataError(self.data_length,self.duid_id,self.df.index[0].date())
            
        #generic first data and sum ==0
        elif (self.df[self.data_col].sum()==0) & self.__first_data():
            print (self.duid_id, self.duid, self.df.index[0]) 
            self.__interpolate_missing_data(d1,d2,quality_flag=0)            

       #First data = BULGANA1     
        elif (self.duid_id==771) & (self.df.index[0]==datetime.datetime(2020,5,12,0,5)):
            print (self.duid_id, self.duid, self.df.index[0]) 
            self.__interpolate_missing_data(d1,d2,quality_flag=0)            
            
        else:
            self.__interpolate_missing_data(d1,d2,quality_flag=1)
            print ("Zero interpolation for {0}".format(self.duid_id))
            if self.df[self.df.QUALITY_FLAG==1].max()[self.data_col] != 0:                
                raise MissingDataError(self.data_length,self.duid_id,self.df.index[0].date())
            
    def __interpolate_missing_data(self,d1,d2,quality_flag):            
        index = pd.date_range(start= d1, end= d2, freq='5min')
        self.df = self.df.reindex(index)
        self.df[self.data_col].interpolate(inplace=True, limit_direction='both')
        self.df['QUALITY_FLAG'].fillna(quality_flag, inplace=True)
    
    def trading_energy_data(self):
        df = pd.DataFrame(    [d for d in self.__trading_energy_generator()], 
                            columns = ['SETTLEMENTDATE', 'DUID','OUTPUT_MWH', 'QUALITY_FLAG' ]    )
        return df
            
    def __trading_energy_generator(self):
        self.df.sort_index(inplace=True)
        date = self.df.index[0].date()
        t_start = datetime.datetime(date.year, date.month,date.day,0,5)
    
        #48 trading intervals in the day 
        #(could be better with groupby function)
        for TI in range(48):
            #t_i initial timestamp of trading_interval, t_f = final timestamp of trading interval
            t_i = t_start + datetime.timedelta(0,1800*TI)
            t_f = t_start + datetime.timedelta(0,1800*(TI+1))
            
            d_ti = self.df[(self.df.index>=t_i) & (self.df.index<=t_f)]    
            #yield d_ti.index[-2], self.duid_id, self.__trapezium_integration(d_ti), d_ti['QUALITY_FLAG'].min()
            yield d_ti.index[-2], self.duid_id, self.__trapezium_integration(d_ti), d_ti['QUALITY_FLAG'].max()
    
    def __trapezium_integration(self,d_ti):
        return 0.5*(d_ti[self.data_col] * [1,2,2,2,2,2,1]).sum()/12        
        
    def __first_data(self):
        sql =    "SELECT MIN(SETTLEMENTDATE) FROM DISPATCH_UNIT_SOLUTION_OLD "\
            "WHERE DUID = {0}".format(self.duid_id)
        cursor = engine.execute(sql)            
        return self.df.index[0]==cursor.fetchall()[0][0]
        

def aggregater(d = datetime.datetime(2010,12,31),db="nemweb"):
    while True:
        df = get_solution(d1=d,db=db)
        try:
            df.data_transaction()
        except DataError as E:
            raise E
            
        print(d)
        d-= datetime.timedelta(1)
        
def nonsched_aggregater(d = datetime.datetime(2016,1,12),db="nemweb"):
    #started at 2016-1-1
    #completed up to at 2016-1-2
    while True:
        df_raw = get_non_schedgen(d1=d,db=db)
        df = DailyNonSched(d1=d,df=df_raw)
        try:
            df.data_transaction()
        except DataError as E:
            raise E
            
        print(d)
        d+= datetime.timedelta(1)

def solution_aggregater(d,db="nemweb_live"):
    #d = "SELECT max(SETTLEMENTDATE) FROM nemweb_derived.TRADING_DUID_OUTPUT"
    d = datetime.datetime(d.year,d.month,d.day)
    while d.date() < datetime.date.today():
        df = get_solution(d1=d,db=db)
        try:
            df.data_transaction()
        except DataError as E:
            raise E        
        d+= datetime.timedelta(1)
        
def trading_aggregator():
    #TODO - check non sched / sched tables aggregation aligns
    cursor = orm.engine_derived.execute("SELECT max(SETTLEMENTDATE) FROM nemweb_derived.TRADING_DUID_OUTPUT")
    solution_aggregater(cursor.fetchall()[0][0],db="nemweb_live")
    non_schedgen_aggregator(cursor.fetchall()[0][0], db="nemweb_live")
            
class NonSchedDUID(DUID):
    """Class for storing a days worth of non-sched duid data"""
    def __init__(self,dataframe,data_col='MWH_READING'):    
        DUID.__init__(self, dataframe,data_col)
        
    def validate_scada_data(self,d1,d2):                
        if self.data_length < 290:
            self.__interpolate_missing_data(d1,d2)
        
    def __interpolate_missing_data(self,d1,d2):            
        index = pd.date_range(start= d1, end= d2, freq='5min')
        self.df = self.df.reindex(index)
        self.df = pd.concat(self.non_sched_yielder(d1=d1,d2=d2))
        self.df.MWH_READING.interpolate(inplace=True)
    
    def non_sched_yielder(self,d1=None,d2=None):
        new_group_flag = self.df.MWH_READING.isna() & self.df.MWH_READING.shift().notna()
        for i,group in self.df.groupby(new_group_flag.cumsum()):                        
            nan_count = sum(group.MWH_READING.isna())
            dx = group.copy()
            if self.df.MWH_READING.max() == 0:
                #Interpolating zero gen
                dx.loc[dx.MWH_READING.isna(),"QUALITY_FLAG"]=1                
            
            elif nan_count <= 6:
                #one trading_interval or less interpolated
                dx.loc[dx.MWH_READING.isna(),"QUALITY_FLAG"]=1                
            elif nan_count <= 24:
                #one more than one trading interval, less than 2 hours interpolated
                dx.loc[dx.MWH_READING.isna(),"QUALITY_FLAG"]=2
            elif nan_count <= 144:
                print ("Warning: {0} nan count = {1}".format(self.duid, nan_count))
                #one more than 2 hours trading interval, less than 12 hours interpolated
                dx.loc[dx.MWH_READING.isna(),"QUALITY_FLAG"]=3
            elif (self.duid == 'GRIFSF1') & (self.df.index[0].date() <= datetime.date(2018,4,20)):                
                #essentially interpolating zero ? Commissioning?
                # first gen for Griffith solar farm
                dx.loc[dx.MWH_READING.isna(),"QUALITY_FLAG"]=1
            else:
                print (nan_count)
                raise MissingDataError(self.data_length,self.duid_id,self.df.index[0].date())            
            yield dx
            

def gc_interpolater(df, test=""): #_test
    #only included current and complete empty days
    dx = df[df.index.date>df.index[0].date()]
    d0 = dx.index[0] 
    dn = d0 + datetime.timedelta(1,300)
    
    while dn <= dx.index[-1]:
        d = dx[(dx.index>=d0) & (dx.index<=dn)]
        duid = NonSchedDUID(d)
        df_t = duid.trading_energy_data()

        df_t.to_sql(name="TRADING_DUID_OUTPUT{}".format(test), index=None, con=engine, if_exists='append')                

        d0 += datetime.timedelta(1)
        dn += datetime.timedelta(1)    
        
def gc_interpolater_one(df, test=""): #_test
    #only first day    
    d0 = df.index[0] 
    dn = d0 + datetime.timedelta(1,300)
    
    
    d = df[(df.index>=d0) & (df.index<=dn)]

    duid = NonSchedDUID(d)
    df_t = duid.trading_energy_data()

    df_t.to_sql(name="TRADING_DUID_OUTPUT{}".format(test), index=None, con=engine, if_exists='append')            

def non_schedgen_aggregator(d_trading, db="nemweb_live"):    
    #d_trading = "SELECT max(SETTLEMENTDATE) FROM nemweb_derived.TRADING_DUID_OUTPUT"
    
    #this needs to be more sophisticates
    #last date metered = YYYY-MM-DD 04:00
    #last date trading = YYYY-MM-DD 00:00  (metered minus 4hrs)    
    
    cursor = orm.engine2.execute("SELECT MAX(INTERVAL_DATETIME) FROM nemweb_live.METER_DATA_GEN_DUID")
    d_metered = cursor.fetchall()[0][0]
    
    d = datetime.datetime(d_trading.year,d_trading.month,d_trading.day)

    while d.date() < d_metered.date():
        print (d)    
        df_ns = get_non_schedgen(d1=d,db=db)
        ns =  DailyNonSched(d,df_ns)
            
        ns.data_transaction()
        d+= datetime.timedelta(1)
        
                        
    
                
def check_duid_range(ds, check="MIN"):
    sql =    "SELECT {0}(SETTLEMENTDATE) FROM DISPATCH_UNIT_SOLUTION_OLD "\
            "WHERE DUID = {1}".format(check,ds.duid_id)

    cursor = engine.execute(sql)
    result = cursor.fetchall()[0][0]
            
    if (ds.index[0]==result) & (check == "MIN"):
        return True            
        
    elif (ds.index[-1]==result) & (check == "MAX"):
        return True
        
    else:
        return False

def range_check():
    ds = datetime.datetime(2014,8,28)
    while True:
        ds -= datetime.timedelta(1)
        print(ds)
        df = select_day_solution(d1=ds)
        data_check(df)        
        
def select_state(regionid=3):
    sql = "SELECT D.TECHFUEL_ID,SUM(T.SCADAVALUE)/12000 as GWH "\
            "FROM nemweb.DISPATCH_UNIT_SCADA T "\
            "INNER JOIN nemweb_live.DUID D "\
            "ON D.ID = T.DUID "\
            "WHERE D.REGIONID = {0} "\
                "AND T.SETTLEMENTDATE > '2017-7-1' "\
            "GROUP BY D.TECHFUEL_ID".format(regionid)

    df = pd.read_sql(sql, con=orm.engine_derived, index_col="TECHFUEL_ID")        
    return df
        
class DataError(ValueError):
    def __str__(self):
        return "{0} data points for DUID {1} ({2})".format(self.args[0],self.args[1],self.args[2])

class MissingDataError(DataError):
    pass
    
class ExtraDataError(DataError):
    pass
    
    

"""
89 MPP_2 2001-08-15 09:35:00
72 CPP_4 2001-07-24 10:00:00
3 BLOWERNG 2001-06-22 11:35:00
21 GUTHEGA 2001-06-22 11:35:00
37 TUMUT3 2001-06-22 11:35:00
38 UPPTUMUT 2001-06-22 11:35:00
231 MURRAY 2001-06-22 11:35:00

"""

"""
interpolated_data"
- 2000-6-26
- 2000-6-16
- 2000-6-13
- 2000-5-15
"""



"""Trading interval indexes

Revision ID: 659f03439631
Revises: 9ee998713b18
Create Date: 2020-10-09 03:01:56.054462

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "659f03439631"
down_revision = "9ee998713b18"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("COMMIT")

    op.create_index(
        "idx_facility_scada_trading_interval_perth_year",
        "facility_scada",
        [sa.text("date_trunc('year', trading_interval AT TIME ZONE 'AWST')")],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_perth_month",
        "facility_scada",
        [sa.text("date_trunc('month', trading_interval AT TIME ZONE 'AWST')")],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_perth_day",
        "facility_scada",
        [sa.text("date_trunc('day', trading_interval AT TIME ZONE 'AWST')")],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_perth_hour",
        "facility_scada",
        [sa.text("date_trunc('hour', trading_interval AT TIME ZONE 'AWST')")],
        postgresql_concurrently=True,
    )

    # wem balancing summary
    op.create_index(
        "idx_balancing_summary_trading_interval_perth_year",
        "balancing_summary",
        [sa.text("date_trunc('year', trading_interval AT TIME ZONE 'AWST')")],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_balancing_summary_trading_interval_perth_month",
        "balancing_summary",
        [sa.text("date_trunc('month', trading_interval AT TIME ZONE 'AWST')")],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_balancing_summary_trading_interval_perth_day",
        "balancing_summary",
        [sa.text("date_trunc('day', trading_interval AT TIME ZONE 'AWST')")],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_balancing_summary_trading_interval_perth_hour",
        "balancing_summary",
        [sa.text("date_trunc('hour', trading_interval AT TIME ZONE 'AWST')")],
        postgresql_concurrently=True,
    )

    # NEM timezone indexes
    # facility scada
    op.create_index(
        "idx_facility_scada_trading_interval_sydney_year",
        "facility_scada",
        [sa.text("date_trunc('year', trading_interval AT TIME ZONE 'AEST')")],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_sydney_month",
        "facility_scada",
        [sa.text("date_trunc('month', trading_interval AT TIME ZONE 'AEST')")],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_sydney_day",
        "facility_scada",
        [sa.text("date_trunc('day', trading_interval AT TIME ZONE 'AEST')")],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_sydney_hour",
        "facility_scada",
        [sa.text("date_trunc('hour', trading_interval AT TIME ZONE 'AEST')")],
        postgresql_concurrently=True,
    )

    # NEM balancing summary
    op.create_index(
        "idx_balancing_summary_trading_interval_sydney_year",
        "balancing_summary",
        [sa.text("date_trunc('year', trading_interval AT TIME ZONE 'AEST')")],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_balancing_summary_trading_interval_sydney_month",
        "balancing_summary",
        [sa.text("date_trunc('month', trading_interval AT TIME ZONE 'AEST')")],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_balancing_summary_trading_interval_sydney_day",
        "balancing_summary",
        [sa.text("date_trunc('day', trading_interval AT TIME ZONE 'AEST')")],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_balancing_summary_trading_interval_sydney_hour",
        "balancing_summary",
        [sa.text("date_trunc('hour', trading_interval AT TIME ZONE 'AEST')")],
        postgresql_concurrently=True,
    )


def downgrade():
    op.drop_index("idx_facility_scada_trading_interval_perth_year")
    op.drop_index("idx_facility_scada_trading_interval_perth_month")
    op.drop_index("idx_facility_scada_trading_interval_perth_day")
    op.drop_index("idx_facility_scada_trading_interval_perth_hour")

    op.drop_index("idx_balancing_summary_trading_interval_perth_year")
    op.drop_index("idx_fbalancing_summary_trading_interval_perth_month")
    op.drop_index("idx_balancing_summary_trading_interval_perth_day")
    op.drop_index("idx_balancing_summary_trading_interval_perth_hour")

    op.drop_index("idx_facility_scada_trading_interval_sydney_year")
    op.drop_index("idx_facility_scada_trading_interval_sydney_month")
    op.drop_index("idx_facility_scada_trading_interval_sydney_day")
    op.drop_index("idx_facility_scada_trading_interval_sydney_hour")

    op.drop_index("idx_balancing_summary_trading_interval_sydney_year")
    op.drop_index("idx_balancing_summary_trading_interval_sydney_month")
    op.drop_index("idx_balancing_summary_trading_interval_sydney_day")
    op.drop_index("idx_balancing_summary_trading_interval_sydney_hour")



import logging
import os
from typing import Dict

from smart_open import open

from opennem.settings import settings

logger = logging.getLogger(__name__)


def get_upload_args(content_type: str = "application/json") -> Dict:
    return {"ContentType": content_type}


def write_to_s3(
    file_path: str, data: str, content_type: str = "application/json"
) -> int:
    """
        Write data to an s3 path
    """
    s3_save_path = os.path.join(settings.s3_bucket_path, file_path)
    write_count = 0
    upload_args = get_upload_args(content_type)

    with open(
        s3_save_path,
        "w",
        transport_params=dict(multipart_upload_kwargs=upload_args),
    ) as fh:
        write_count = fh.write(data)

    logger.info("Wrote {} to {}".format(len(data), s3_save_path))

    return write_count



"""

NEMWEB Data ingress into OpenNEM format


"""

import logging
from datetime import datetime
from typing import Optional

from sqlalchemy.dialects.postgresql import insert

from opennem.core.networks import NetworkNEM
from opennem.core.normalizers import clean_float, normalize_duid
from opennem.db import SessionLocal, get_database_engine
from opennem.db.models.opennem import BalancingSummary, FacilityScada
from opennem.schema.network import NetworkSchema
from opennem.utils.dates import parse_date
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


def process_case_solutions(table):
    pass


def process_pre_ap_price(table, spider):
    session = SessionLocal()
    engine = get_database_engine()

    if "records" not in table:
        raise Exception("Invalid table no records")

    records = table["records"]

    records_to_store = []

    for record in records:
        trading_interval = parse_date(
            record["SETTLEMENTDATE"], network=NetworkNEM, dayfirst=False
        )

        if not trading_interval:
            continue

        records_to_store.append(
            {
                "network_id": "NEM",
                "created_by": spider.name,
                # "updated_by": None,
                "network_region": record["REGIONID"],
                "trading_interval": trading_interval,
                "price": record["PRE_AP_ENERGY_PRICE"],
            }
        )

    stmt = insert(BalancingSummary).values(records_to_store)
    stmt.bind = engine
    stmt = stmt.on_conflict_do_update(
        index_elements=["trading_interval", "network_id", "network_region"],
        set_={"price": stmt.excluded.price},
    )

    try:
        session.execute(stmt)
        session.commit()
    except Exception as e:
        logger.error("Error inserting records")
        logger.error(e)
        return 0
    finally:
        session.close()

    return len(records_to_store)


def unit_scada_generate_facility_scada(
    records,
    spider=None,
    network: NetworkSchema = NetworkNEM,
    interval_field: str = "SETTLEMENTDATE",
    facility_code_field: str = "DUID",
    power_field: Optional[str] = None,
    energy_field: Optional[str] = None,
):
    created_at = datetime.now()
    primary_keys = []

    for row in records:

        trading_interval = parse_date(
            row[interval_field], network=network, dayfirst=False
        )

        if facility_code_field not in row:
            logger.error("Invalid row no facility_code")
            continue

        facility_code = normalize_duid(row[facility_code_field])

        pkey = (trading_interval, facility_code)

        if pkey in primary_keys:
            continue

        primary_keys.append(pkey)

        generated = None

        if power_field and power_field in row:
            generated = clean_float(row[power_field])

        energy = None

        if energy_field and energy_field in row:
            energy = clean_float(row[energy_field])

        created_by = ""

        if spider and hasattr(spider, "name"):
            created_by = spider.name

        yield {
            "created_by": created_by,
            "created_at": created_at,
            "updated_at": None,
            "network_id": "WEM",
            "trading_interval": trading_interval,
            "facility_code": facility_code,
            "generated": generated,
            "eoi_quantity": energy,
        }


def process_unit_scada(table, spider):
    if "records" not in table:
        raise Exception("Invalid table no records")

    records = table["records"]
    item = dict()

    item["table_schema"] = FacilityScada
    item["update_fields"] = ["generated"]
    item["records"] = unit_scada_generate_facility_scada(
        records, spider, power_field="SCADAVALUE"
    )
    item["content"] = None

    return item


def process_unit_solution(table, spider):
    if "records" not in table:
        raise Exception("Invalid table no records")

    records = table["records"]
    item = dict()

    item["table_schema"] = FacilityScada
    item["update_fields"] = ["generated"]
    item["records"] = unit_scada_generate_facility_scada(
        records,
        spider,
        interval_field="SETTLEMENTDATE",
        facility_code_field="DUID",
        power_field="INITIALMW",
    )
    item["content"] = None

    return item


def process_meter_data_gen_duid(table, spider):
    if "records" not in table:
        raise Exception("Invalid table no records")

    records = table["records"]
    item = dict()

    item["table_schema"] = FacilityScada
    item["update_fields"] = ["eoi_generated"]
    item["records"] = unit_scada_generate_facility_scada(
        records,
        spider,
        interval_field="INTERVAL_DATETIME",
        facility_code_field="DUID",
        energy_field="MWH_READING",
    )
    item["content"] = None

    return item


TABLE_PROCESSOR_MAP = {
    "METER_DATA_GEN_DUID": "process_meter_data_gen_duid",
    "DISPATCH_CASE_SOLUTION": "process_case_solutions",
    "DISPATCH_UNIT_SCADA": "process_unit_scada",
    "DISPATCH_UNIT_SOLUTION": "process_unit_solution",
    "DISPATCH_PRE_AP_PRICE": "process_pre_ap_price",
}


class NemwebUnitScadaOpenNEMStorePipeline(object):
    @check_spider_pipeline
    def process_item(self, item, spider=None):

        if "tables" not in item:
            logger.error("Invalid item - no tables located")
            return 0

        if not isinstance(item["tables"], list):
            logger.error("Invalid item - no tables located")
            return 0

        tables = item["tables"]

        ret = 0

        for table in tables:
            if "name" not in table:
                logger.info("Invalid table found")
                continue

            table_name = table["name"]

            if table_name not in TABLE_PROCESSOR_MAP:
                logger.error("No processor for table %s", table_name)
                continue

            process_meth = TABLE_PROCESSOR_MAP[table_name]

            if process_meth not in globals():
                logger.info("Invalid processing function %s", process_meth)
                continue

            ret = globals()[process_meth](table, spider=spider)

        return ret



"""
Timezone-related classes and functions.

Some methods adapted from the Django project
"""
from datetime import datetime, timedelta, timezone

import pytz

UTC = pytz.utc


def get_current_timezone():
    return UTC


def get_fixed_timezone(offset):
    """Return a tzinfo instance with a fixed offset from UTC."""

    if isinstance(offset, timedelta):
        offset = offset.total_seconds() // 60

    sign = "-" if offset < 0 else "+"
    hhmm = "%02d%02d" % divmod(abs(offset), 60)
    name = sign + hhmm

    return timezone(timedelta(minutes=offset), name)


def is_aware(value: datetime) -> bool:
    """
    Determines if a given datetime.datetime is aware.

    The concept is defined in Python's docs:
    http://docs.python.org/library/datetime.html#datetime.tzinfo

    Assuming value.tzinfo is either None or a proper datetime.tzinfo,
    value.utcoffset() implements the appropriate logic.
    """

    return value.utcoffset() is not None


def is_naive(value: datetime) -> bool:
    """
    Determines if a given datetime.datetime is naive.

    The concept is defined in Python's docs:
    http://docs.python.org/library/datetime.html#datetime.tzinfo

    Assuming value.tzinfo is either None or a proper datetime.tzinfo,
    value.utcoffset() implements the appropriate logic.
    """

    return value.utcoffset() is None


def make_aware(value, timezone=None, is_dst=None):
    """Make a naive datetime.datetime in a given time zone aware."""

    if timezone is None:
        timezone = get_current_timezone()

    if hasattr(timezone, "localize"):
        # This method is available for pytz time zones.
        return timezone.localize(value, is_dst=is_dst)
    else:
        # Check that we won't overwrite the timezone of an aware datetime.
        if is_aware(value):
            raise ValueError(
                "make_aware expects a naive datetime, got %s" % value
            )
        # This may be wrong around DST changes!
        return value.replace(tzinfo=timezone)


def make_naive(value, timezone=None):
    """Make an aware datetime.datetime naive in a given time zone."""

    if timezone is None:
        timezone = get_current_timezone()

    # Emulate the behavior of astimezone() on Python < 3.6.
    if is_naive(value):
        raise ValueError("make_naive() cannot be applied to a naive datetime")

    return value.astimezone(timezone).replace(tzinfo=None)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['opennem',
 'opennem.api',
 'opennem.api.admin',
 'opennem.api.export',
 'opennem.api.facility',
 'opennem.api.geo',
 'opennem.api.photo',
 'opennem.api.revision',
 'opennem.api.station',
 'opennem.api.stats',
 'opennem.api.weather',
 'opennem.commands',
 'opennem.core',
 'opennem.core.facility',
 'opennem.core.stations',
 'opennem.core.unit',
 'opennem.db',
 'opennem.db.migrations',
 'opennem.db.migrations.versions',
 'opennem.db.models',
 'opennem.diff',
 'opennem.exporter',
 'opennem.geo',
 'opennem.importer',
 'opennem.nem_derived',
 'opennem.pipelines',
 'opennem.pipelines.aemo',
 'opennem.pipelines.aemo.mms',
 'opennem.pipelines.apvi',
 'opennem.pipelines.nem',
 'opennem.pipelines.npi',
 'opennem.pipelines.wem',
 'opennem.scheduler',
 'opennem.schema',
 'opennem.settings',
 'opennem.spiders',
 'opennem.spiders.aemo',
 'opennem.spiders.apvi',
 'opennem.spiders.bom',
 'opennem.spiders.nem',
 'opennem.spiders.npi',
 'opennem.spiders.wem',
 'opennem.utils']

package_data = \
{'': ['*'],
 'opennem': ['data/bom_capitals.json',
             'data/bom_capitals.json',
             'data/bom_capitals.json',
             'data/bom_capitals.json',
             'data/bom_capitals.json',
             'data/bom_capitals.json',
             'data/bom_capitals.json',
             'data/bom_capitals.json',
             'data/bom_capitals.json',
             'data/bom_capitals.json',
             'data/bom_capitals.json',
             'data/bom_capitals.json',
             'data/bom_capitals.json',
             'data/bom_capitals.json',
             'data/bom_capitals.json',
             'data/bom_capitals.json',
             'data/bom_stations.json',
             'data/bom_stations.json',
             'data/bom_stations.json',
             'data/bom_stations.json',
             'data/bom_stations.json',
             'data/bom_stations.json',
             'data/bom_stations.json',
             'data/bom_stations.json',
             'data/bom_stations.json',
             'data/bom_stations.json',
             'data/bom_stations.json',
             'data/bom_stations.json',
             'data/bom_stations.json',
             'data/bom_stations.json',
             'data/bom_stations.json',
             'data/bom_stations.json',
             'data/diff_report.md',
             'data/diff_report.md',
             'data/diff_report.md',
             'data/diff_report.md',
             'data/diff_report.md',
             'data/diff_report.md',
             'data/diff_report.md',
             'data/diff_report.md',
             'data/diff_report.md',
             'data/diff_report.md',
             'data/diff_report.md',
             'data/diff_report.md',
             'data/diff_report.md',
             'data/diff_report.md',
             'data/diff_report.md',
             'data/diff_report.md',
             'data/mms.json',
             'data/mms.json',
             'data/mms.json',
             'data/mms.json',
             'data/mms.json',
             'data/mms.json',
             'data/mms.json',
             'data/mms.json',
             'data/mms.json',
             'data/mms.json',
             'data/mms.json',
             'data/mms.json',
             'data/mms.json',
             'data/mms.json',
             'data/mms.json',
             'data/mms.json',
             'data/mms_duid_station_map.json',
             'data/mms_duid_station_map.json',
             'data/mms_duid_station_map.json',
             'data/mms_duid_station_map.json',
             'data/mms_duid_station_map.json',
             'data/mms_duid_station_map.json',
             'data/mms_duid_station_map.json',
             'data/mms_duid_station_map.json',
             'data/mms_duid_station_map.json',
             'data/mms_duid_station_map.json',
             'data/mms_duid_station_map.json',
             'data/mms_duid_station_map.json',
             'data/mms_duid_station_map.json',
             'data/mms_duid_station_map.json',
             'data/mms_duid_station_map.json',
             'data/mms_duid_station_map.json',
             'data/nem_gi.json',
             'data/nem_gi.json',
             'data/nem_gi.json',
             'data/nem_gi.json',
             'data/nem_gi.json',
             'data/nem_gi.json',
             'data/nem_gi.json',
             'data/nem_gi.json',
             'data/nem_gi.json',
             'data/nem_gi.json',
             'data/nem_gi.json',
             'data/nem_gi.json',
             'data/nem_gi.json',
             'data/nem_gi.json',
             'data/nem_gi.json',
             'data/nem_gi.json',
             'data/npi_facilities.json',
             'data/npi_facilities.json',
             'data/npi_facilities.json',
             'data/npi_facilities.json',
             'data/npi_facilities.json',
             'data/npi_facilities.json',
             'data/npi_facilities.json',
             'data/npi_facilities.json',
             'data/npi_facilities.json',
             'data/npi_facilities.json',
             'data/npi_facilities.json',
             'data/npi_facilities.json',
             'data/npi_facilities.json',
             'data/npi_facilities.json',
             'data/npi_facilities.json',
             'data/npi_facilities.json',
             'data/opennem.json',
             'data/opennem.json',
             'data/opennem.json',
             'data/opennem.json',
             'data/opennem.json',
             'data/opennem.json',
             'data/opennem.json',
             'data/opennem.json',
             'data/opennem.json',
             'data/opennem.json',
             'data/opennem.json',
             'data/opennem.json',
             'data/opennem.json',
             'data/opennem.json',
             'data/opennem.json',
             'data/opennem.json',
             'data/registry.json',
             'data/registry.json',
             'data/registry.json',
             'data/registry.json',
             'data/registry.json',
             'data/registry.json',
             'data/registry.json',
             'data/registry.json',
             'data/registry.json',
             'data/registry.json',
             'data/registry.json',
             'data/registry.json',
             'data/registry.json',
             'data/registry.json',
             'data/registry.json',
             'data/registry.json',
             'data/rel.json',
             'data/rel.json',
             'data/rel.json',
             'data/rel.json',
             'data/rel.json',
             'data/rel.json',
             'data/rel.json',
             'data/rel.json',
             'data/rel.json',
             'data/rel.json',
             'data/rel.json',
             'data/rel.json',
             'data/rel.json',
             'data/rel.json',
             'data/rel.json',
             'data/rel.json',
             'data/stations.csv',
             'data/stations.csv',
             'data/stations.csv',
             'data/stations.csv',
             'data/stations.csv',
             'data/stations.csv',
             'data/stations.csv',
             'data/stations.csv',
             'data/stations.csv',
             'data/stations.csv',
             'data/stations.csv',
             'data/stations.csv',
             'data/stations.csv',
             'data/stations.csv',
             'data/stations.csv',
             'data/stations.csv',
             'data/stations.geojson',
             'data/stations.geojson',
             'data/stations.geojson',
             'data/stations.geojson',
             'data/stations.geojson',
             'data/stations.geojson',
             'data/stations.geojson',
             'data/stations.geojson',
             'data/stations.geojson',
             'data/stations.geojson',
             'data/stations.geojson',
             'data/stations.geojson',
             'data/stations.geojson',
             'data/stations.geojson',
             'data/stations.geojson',
             'data/stations.geojson',
             'data/stations.json',
             'data/stations.json',
             'data/stations.json',
             'data/stations.json',
             'data/stations.json',
             'data/stations.json',
             'data/stations.json',
             'data/stations.json',
             'data/stations.json',
             'data/stations.json',
             'data/stations.json',
             'data/stations.json',
             'data/stations.json',
             'data/stations.json',
             'data/stations.json',
             'data/stations.json',
             'data/wikidata-parsed.json',
             'data/wikidata-parsed.json',
             'data/wikidata-parsed.json',
             'data/wikidata-parsed.json',
             'data/wikidata-parsed.json',
             'data/wikidata-parsed.json',
             'data/wikidata-parsed.json',
             'data/wikidata-parsed.json',
             'data/wikidata-parsed.json',
             'data/wikidata-parsed.json',
             'data/wikidata-parsed.json',
             'data/wikidata-parsed.json',
             'data/wikidata-parsed.json',
             'data/wikidata-parsed.json',
             'data/wikidata-parsed.json',
             'data/wikidata-parsed.json',
             'data/wikidata-photos.json',
             'data/wikidata-photos.json',
             'data/wikidata-photos.json',
             'data/wikidata-photos.json',
             'data/wikidata-photos.json',
             'data/wikidata-photos.json',
             'data/wikidata-photos.json',
             'data/wikidata-photos.json',
             'data/wikidata-photos.json',
             'data/wikidata-photos.json',
             'data/wikidata-photos.json',
             'data/wikidata-photos.json',
             'data/wikidata-photos.json',
             'data/wikidata-photos.json',
             'data/wikidata-photos.json',
             'data/wikidata-photos.json',
             'data/wikidata.json',
             'data/wikidata.json',
             'data/wikidata.json',
             'data/wikidata.json',
             'data/wikidata.json',
             'data/wikidata.json',
             'data/wikidata.json',
             'data/wikidata.json',
             'data/wikidata.json',
             'data/wikidata.json',
             'data/wikidata.json',
             'data/wikidata.json',
             'data/wikidata.json',
             'data/wikidata.json',
             'data/wikidata.json',
             'data/wikidata.json'],
 'opennem.core': ['data/*'],
 'opennem.db': ['fixtures/*', 'views/*']}

install_requires = \
['Pint>=0.16.1,<0.17.0',
 'Wikidata>=0.7.0,<0.8.0',
 'alembic>=1.4.2,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'dictalchemy>=0.1.2,<0.2.0',
 'geoalchemy2>=0.8.4,<0.9.0',
 'geojson-pydantic>=0.2.1,<0.3.0',
 'geojson>=2.5.0,<3.0.0',
 'huey>=2.2.0,<3.0.0',
 'openpyxl>=3.0.4,<4.0.0',
 'pillow>=7.2.0,<8.0.0',
 'pydantic>=1.5.1,<2.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'python-dotenv>=0.14.0,<0.15.0',
 'pytz>=2020.1,<2021.0',
 'pyyaml>=5.3.1,<6.0.0',
 'redis>=3.5.3,<4.0.0',
 'requests>=2.23.0,<3.0.0',
 'requests_cache>=0.5.2,<0.6.0',
 'scrapy>=2.1.0,<3.0.0',
 'sentry-sdk>=0.14.4,<0.15.0',
 'shapely>=1.7.0,<2.0.0',
 'smart_open>=2.0.0,<3.0.0',
 'sqlalchemy>=1.3.17,<2.0.0',
 'wikipedia>=1.4.0,<2.0.0']

extras_require = \
{':extra == "postgres"': ['psycopg2-binary>=2.8.5,<3.0.0'],
 ':extra == "server"': ['uvicorn>=0.11.7,<0.12.0',
                        'fastapi[all]>=0.60.1,<0.61.0']}

entry_points = \
{'console_scripts': ['opennem = opennem.cli:main']}

setup_kwargs = {
    'name': 'opennem',
    'version': '3.0.0a39',
    'description': 'opennem engine agent',
    'long_description': "# OpenNEM Energy Market Data Access\n\nThe OpenNEM project aims to make the wealth of public National Electricity Market (NEM) data more accessible to a wider audience.\n\nThis toolkit enables downloading, mirroring and accessing energy data from various networks\n\nProject homepage at https://opennem.org.au\n\nAvailable on Docker at https://hub.docker.com/r/opennem/opennem\n\nCurrently supports:\n\n- Australian NEM: https://www.nemweb.com.au/\n- West Australia Energy Market: http://data.wa.aemo.com.au/\n\n## Requirements\n\n * Python 3.7+ (see `.python-version` with `pyenv`)\n * Docker and `docker-compose` if you want to run the local dev stack\n\n## Quickstart\n\nWith poetry:\n\n```sh\n$ poetry install\n$ source .venv/bin/activate\n$ ./init.sh\n```\n\nWith pip + venv:\n\n```sh\n$ pip -m venv .venv\n$ pip install -r requirements.txt\n$ source .venv/bin/activate\n$ ./init.sh\n```\n\n## Install\n\nYou can install this project with python `pip`:\n\n```sh\n$ pip install opennem\n```\n\nOr alternatively with docker:\n\n```\n$ docker pull opennem/opennem\n```\n\nBundled with sqlite support. Other database drivers are optional and not installed by default. Install a supported database driver:\n\nPostgres:\n\n```sh\n$ pip install psycopg2\n```\n\n## Install Extras\n\nThe package contains extra modules that can be installed:\n\n```sh\n$ poetry install -E postgres\n```\n\nThe list of extras are:\n\n * `postgres` - Postgres database drivers\n * `server` - API server\n\n## Usage\n\nList the crawlers\n\n```sh\n$ scrapy list\n```\n\nCrawl\n\n```sh\n$ scrapy crawl au.nem.current.dispatch_scada\n```\n\n## Development\n\nThis project uses the new `pyproject.toml` project and build specification file. To make use of it use the `poetry` tool which can be installed on Windows, MacOS and Linux:\n\nhttps://python-poetry.org/docs/\n\nInstallation instructions for Poetry are at:\n\nhttps://python-poetry.org/docs/#installation\n\nBy default poetry will install virtual environments in your home metadata directory. A good alternative is to install the `venv` locally for each project with the following setting:\n\n```sh\n$ poetry config virtualenvs.in-project true\n```\n\nThis will create the virtual environment within the project folder in a folder called `.venv`. This folder is ignored by git by default.\n\nSetting up a virtual environment and installing requiements using Poetry:\n\n```sh\n$ poetry install\n```\n\nTo activate the virtual environment either run:\n\n```sh\n$ poetry shell\n```\n\nOr you can just activate the standard `venv`\n\n```sh\n$ source .venv/bin/activate\n```\n\nSettings are read from environment variables. Environment variables can be read from a `.env` file in the root of the folder. Setup the environment by copying the `.env.sample` file to `.env`. The defaults in the sample file map to the settings in `docker-compose.yml`\n\nThere is a `docker-compose` file that will bring a local database:\n\n```sh\n$ docker-compose up -d\n```\n\nBring up the database migrations using alembic:\n\n```sh\n$ alembic upgrade head\n```\n\nRun scrapy in the root folder for options:\n\n```sh\n$ scrapy\n```\n\nThe `opennem` cli interface provides other options and settings:\n\n```sh\n$ opennem -h\n```\n\nSettings for Visual Studio Code are stored in `.vscode`. Code is kept formatted and linted using `pylint`, `black` and `isort` with settings defined in `pyproject.toml`\n\n## Testing\n\nTests are in `tests/`\n\nRun tests with:\n\n```sh\n$ pytest\n```\n\nRun background test watcher with\n\n```sh\n$ ptw\n```\n\n## Build Release\n\nThe script `build-release.sh` will tag a new release, build the docker image, tag the git version, push to GitHub and push the latest\nrelease to PyPi\n\n\n## Architecture overview\n\nThis project uses [Scrapy](https://scrapy.org/) to obtain data from supported energy markets and [SQLAlchemy](https://www.sqlalchemy.org/) to store data, and [Alembic](https://alembic.sqlalchemy.org/en/latest/) for database migrations. Database storage has been tested with sqlite, postgres and mysql.\n\nOverview of scrapy architecture:\n\n![](https://docs.scrapy.org/en/latest/_images/scrapy_architecture_02.png)\n\n## Code Navigation\n\n* Spider definitions in `opennem/spiders`\n* Processing pipelines for crawls in `opennem/pipelines`\n* Database models for supported energy markets are stored in `opennem/db/models`\n\n## Deploy Crawlers\n\nYou can deploy the crawlers to the scrapyd server with:\n\n```sh\n$ scrapyd-deploy\n```\n\nIf you don't have that command and it isn't available install it with:\n\n```sh\n$ pip install scrapyd-client\n```\n\nWhich installs the [scrapyd-client](https://github.com/scrapy/scrapyd-client) tools. Project settings are read from `scrapy.cfg`\n",
    'author': 'Dylan McConnell',
    'author_email': 'dylan.mcconnell@unimelb.edu.au',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://opennem.org.au',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import functools
import logging
import re

import six


def regex(x):
    if isinstance(x, six.string_types):
        return re.compile(x)
    return x


def check_spider_pipeline(process_item_method):
    @functools.wraps(process_item_method)
    def wrapper(self, item, spider):

        # message template for debugging
        msg = "%%s %s pipeline step" % (self.__class__.__name__,)

        pipelines = set([])

        if hasattr(spider, "pipelines"):
            if type(spider.pipelines) is set:
                pipelines |= spider.pipelines

        if hasattr(spider, "pipelines_extra"):
            if type(spider.pipelines_extra) is set:
                pipelines |= spider.pipelines_extra

        if self.__class__ in pipelines:
            spider.log(msg % "executing", level=logging.DEBUG)
            return process_item_method(self, item, spider)

        else:
            # spider.log(msg % "skipping", level=logging.DEBUG)
            return item

    return wrapper



import json
import logging
from datetime import datetime
from typing import List

from opennem.core.facility.fueltechs import parse_facility_fueltech
from opennem.core.facilitystatus import parse_facility_status
from opennem.core.loader import load_data
from opennem.core.networks import network_from_network_region
from opennem.exporter.encoders import OpenNEMJSONEncoder
from opennem.importer.compat import (
    map_compat_facility_state,
    map_compat_fueltech,
    map_compat_network_region,
)
from opennem.schema.opennem import (
    FacilitySchema,
    LocationSchema,
    StationSchema,
)
from opennem.schema.stations import StationSet

logger = logging.getLogger("opennem.importer.registry")


def registry_to_station(registry_station: dict, _id: int) -> StationSchema:
    station = StationSchema(
        id=_id,
        code=registry_station.get("station_id", ""),
        name=registry_station.get("display_name", ""),
        network_name=registry_station.get("display_name", ""),
    )

    station.location = LocationSchema(
        state=registry_station.get("location", {}).get("state", None),
        geocode_by="opennem.registry",
        geocode_approved=True,
        lat=registry_station.get("location", {}).get("latitude", None),
        lng=registry_station.get("location", {}).get("longitude", None),
    )

    if "duid_data" not in registry_station:
        logger.info(
            "Registry: station has no duid data: {}".format(
                registry_station.get("display_name", "")
            )
        )
        return station

    for duid, registry_facility in registry_station["duid_data"].items():
        network_region = map_compat_network_region(
            registry_station.get("region_id", "")
        )

        facility = FacilitySchema(
            **{
                "code": duid,
                "created_by": "opennem.registry",
                "created_at": datetime.now(),
                "network": network_from_network_region(network_region),
                "network_region": network_region,
                "station_code": registry_station.get("station_id", ""),
                "dispatch_type": "GENERATOR",
                "status": parse_facility_status(
                    map_compat_facility_state(
                        registry_station.get("status", {}).get("state", "")
                    )
                ),
                "fueltech": parse_facility_fueltech(
                    map_compat_fueltech(
                        registry_facility.get("fuel_tech", None)
                    )
                ),
                "capacity_registered": registry_facility.get(
                    "registered_capacity", None
                ),
            }
        )
        station.facilities.append(facility)

    return station


def registry_to_stations(registry, start_id: int = 5000) -> StationSet:

    _id = start_id

    stations = StationSet()

    for station_code, station in registry.items():
        _station = registry_to_station(station, _id)

        if _station:
            station_existing = stations.get_code(_station.code)

            if station_existing:
                logger.info(
                    "Merging station {} into {}".format(
                        station_code, station_existing.code
                    )
                )
                station_existing.facilities += _station.facilities
                continue

            stations.add(_station)
            _id += 1

    return stations


def registry_import() -> StationSet:
    registry = load_data("facility_registry.json")

    on_registry = registry_to_stations(registry)

    return on_registry


def registry_export():
    stations = registry_import()

    with open("data/registry.json", "w") as fh:
        fh.write(stations.json(indent=4))

    logger.info("Wrote {} records".format(stations.length))


if __name__ == "__main__":
    registry_export()



"""
    Parse current and v3 facilities registries and take a diff.


"""

import csv
import json

# from opennem.utils.log_config
import logging
import re
from datetime import timedelta
from itertools import chain
from operator import itemgetter
from pprint import pprint
from typing import List

from mdutils.mdutils import MdUtils

from opennem.core.loader import load_data

from .loader import load_current, load_registry

logger = logging.getLogger("opennem.diff")

markdown_report = []


def report_changed_names(registry, current):
    renames = []

    for station in registry:
        in_current = list(filter(lambda x: x.code == station.code, current))

        if in_current and len(in_current):
            station_current = in_current[0]
            if station_current.name != station.name:
                renames.append(
                    "`{}` renamed to `{}`".format(
                        station.name, station_current.name
                    )
                )

    return renames


def report_changed_fueltechs(registry_units, current_units):
    renames = []

    for unit in registry_units:
        in_current = list(filter(lambda x: x.duid == unit.duid, current_units))

        if in_current and len(in_current):
            unit_current = in_current[0]
            if unit_current.fueltech != unit.fueltech:
                renames.append(
                    "{} (`{}`) fueltech `{}` changed to `{}`".format(
                        unit_current.name,
                        unit_current.duid,
                        unit_current.fueltech,
                        unit.fueltech,
                    )
                )

    return renames


def report_changed_capacities(registry_units, current_units):
    renames = []

    for unit in registry_units:
        in_current = list(filter(lambda x: x.duid == unit.duid, current_units))

        if in_current and len(in_current):
            unit_current = in_current[0]
            if unit_current.capacity == None and unit.capacity:
                renames.append(
                    "{} (`{}`) added capacity `{}`".format(
                        unit_current.name, unit_current.duid, unit.capacity,
                    )
                )
            elif (
                unit_current.capacity
                and unit.capacity
                and round(unit_current.capacity, 0) != round(unit.capacity, 0)
            ):
                renames.append(
                    "{} (`{}`) capacity `{}` changed to `{}`".format(
                        unit_current.name,
                        unit_current.duid,
                        unit_current.capacity,
                        unit.capacity,
                    )
                )

    return renames


def records_diff(
    subject: List[object], comparitor: List[object], key: str = "code"
) -> List[str]:
    diff_keys = list(
        set([getattr(i, key) for i in subject])
        - set([getattr(i, key) for i in comparitor])
    )

    diff_records = list(
        filter(lambda x: getattr(x, key) in diff_keys, subject)
    )

    return diff_records


def station_in_registry_not_in_new(registry, current):
    diff_stations = records_diff(registry, current)

    list_table = ["Name", "Code", "Facilities"]

    [
        list_table.extend([i.name, i.code, str(len(i.facilities))])
        for i in diff_stations
    ]

    return list_table


def stations_in_new_not_in_registry(registry, current):
    diff_stations = records_diff(current, registry)

    list_table = ["Name", "Code", "Facilities"]

    [
        list_table.extend(
            [i.name, str(i.code) if i.code else "", str(len(i.facilities))]
        )
        for i in diff_stations
    ]

    return list_table


get_num_rows = lambda records, columns: int(len(records) / columns)


def run_diff():
    diff_report = {}

    registry = load_registry()
    current = load_current()

    nem_mms = load_data("mms.json", True)
    nem_gi = load_data("nem_gi.json", True)
    nem_rel = load_data("rel.json", True)

    registry_units = list(chain(*[s.facilities for s in registry]))
    current_units = list(chain(*[s.facilities for s in current]))

    registry_station_codes = [station.code for station in registry]
    current_station_codes = [station.code for station in current]

    registry_station_names = [station.name for station in registry]
    current_station_names = [station.name for station in current]

    new_stations = list(
        set(
            [
                station.name
                for station in current
                if station.name not in registry_station_names
                and station.code not in registry_station_codes
            ]
        )
    )

    md = MdUtils(file_name="data/diff_report.md", title="Opennem Report")
    md.new_header(level=1, title="Opennem Report")

    # Summary
    md.new_header(level=1, title="Summary")
    summary = ["", "Prod", "Version 3"]
    summary.extend(["Stations", str(len(registry)), str(len(current))])
    summary.extend(
        ["Units", str(len(registry_units)), str(len(current_units))]
    )
    md.new_table(columns=3, rows=3, text=summary)

    # Renames
    md.new_header(level=1, title="Renamed Stations")
    renames = report_changed_names(registry, current)
    md.new_list(renames)

    # Fueltechs
    md.new_header(level=1, title="Changed Fueltechs")
    renames = report_changed_fueltechs(registry_units, current_units)
    md.new_list(renames)

    # Capacities
    md.new_header(level=1, title="Changed Capacities")
    renames = report_changed_capacities(registry_units, current_units)
    md.new_list(renames)

    # Old stations not in new
    md.new_header(level=1, title="Stations not in current")
    not_in_news = station_in_registry_not_in_new(registry, current)
    md.new_table(
        columns=3,
        rows=get_num_rows(not_in_news, 3),
        text=not_in_news,
        text_align="left",
    )

    # Old stations not in new
    md.new_header(level=1, title="New Stations")
    not_in_registry = stations_in_new_not_in_registry(registry, current)
    md.new_table(
        columns=3,
        rows=get_num_rows(not_in_registry, 3),
        text=not_in_registry,
        text_align="left",
    )

    md.new_table_of_contents(table_title="Contents", depth=2)
    md.create_md_file()


if __name__ == "__main__":
    run_diff()
